from veros.core.operators import numpy as npx
from veros import logger

from veros import veros_kernel, veros_routine, KernelOutput
from veros.variables import allocate
from veros.core import density, utilities
from veros.core.operators import update, update_add, at


@veros_kernel
def dm_taper(sx, iso_slopec, iso_dslope):
    """
    tapering function for isopycnal slopes
    """
    return 0.5 * (1.0 + npx.tanh((-npx.abs(sx) + iso_slopec) / iso_dslope))


@veros_kernel
def isoneutral_diffusion_pre(state):
    """
    Isopycnal diffusion for tracer
    following functional formulation by Griffies et al
    Code adopted from MOM2.1
    """
    vs = state.variables
    settings = state.settings

    epsln = 1e-20

    dTdx = allocate(state.dimensions, ("xt", "yt", "zt"))
    dSdx = allocate(state.dimensions, ("xt", "yt", "zt"))
    dTdy = allocate(state.dimensions, ("xt", "yt", "zt"))
    dSdy = allocate(state.dimensions, ("xt", "yt", "zt"))
    dTdz = allocate(state.dimensions, ("xt", "yt", "zt"))
    dSdz = allocate(state.dimensions, ("xt", "yt", "zt"))

    """
    drho_dt and drho_ds at centers of T cells
    """
    drdT = vs.maskT * density.get_drhodT(state, vs.salt[:, :, :, vs.tau], vs.temp[:, :, :, vs.tau], npx.abs(vs.zt))
    drdS = vs.maskT * density.get_drhodS(state, vs.salt[:, :, :, vs.tau], vs.temp[:, :, :, vs.tau], npx.abs(vs.zt))

    """
    gradients at top face of T cells
    """
    dTdz = update(
        dTdz,
        at[:, :, :-1],
        vs.maskW[:, :, :-1]
        * (vs.temp[:, :, 1:, vs.tau] - vs.temp[:, :, :-1, vs.tau])
        / vs.dzw[npx.newaxis, npx.newaxis, :-1],
    )
    dSdz = update(
        dSdz,
        at[:, :, :-1],
        vs.maskW[:, :, :-1]
        * (vs.salt[:, :, 1:, vs.tau] - vs.salt[:, :, :-1, vs.tau])
        / vs.dzw[npx.newaxis, npx.newaxis, :-1],
    )

    """
    gradients at eastern face of T cells
    """
    dTdx = update(
        dTdx,
        at[:-1, :, :],
        vs.maskU[:-1, :, :]
        * (vs.temp[1:, :, :, vs.tau] - vs.temp[:-1, :, :, vs.tau])
        / (vs.dxu[:-1, npx.newaxis, npx.newaxis] * vs.cost[npx.newaxis, :, npx.newaxis]),
    )
    dSdx = update(
        dSdx,
        at[:-1, :, :],
        vs.maskU[:-1, :, :]
        * (vs.salt[1:, :, :, vs.tau] - vs.salt[:-1, :, :, vs.tau])
        / (vs.dxu[:-1, npx.newaxis, npx.newaxis] * vs.cost[npx.newaxis, :, npx.newaxis]),
    )

    """
    gradients at northern face of T cells
    """
    dTdy = update(
        dTdy,
        at[:, :-1, :],
        vs.maskV[:, :-1, :]
        * (vs.temp[:, 1:, :, vs.tau] - vs.temp[:, :-1, :, vs.tau])
        / vs.dyu[npx.newaxis, :-1, npx.newaxis],
    )
    dSdy = update(
        dSdy,
        at[:, :-1, :],
        vs.maskV[:, :-1, :]
        * (vs.salt[:, 1:, :, vs.tau] - vs.salt[:, :-1, :, vs.tau])
        / vs.dyu[npx.newaxis, :-1, npx.newaxis],
    )

    """
    Compute Ai_ez and K11 on center of east face of T cell.
    """
    diffloc = allocate(state.dimensions, ("xt", "yt", "zt"))
    diffloc = update(
        diffloc,
        at[1:-2, 2:-2, 1:],
        0.25
        * (vs.K_iso[1:-2, 2:-2, 1:] + vs.K_iso[1:-2, 2:-2, :-1] + vs.K_iso[2:-1, 2:-2, 1:] + vs.K_iso[2:-1, 2:-2, :-1]),
    )
    diffloc = update(diffloc, at[1:-2, 2:-2, 0], 0.5 * (vs.K_iso[1:-2, 2:-2, 0] + vs.K_iso[2:-1, 2:-2, 0]))

    sumz = allocate(state.dimensions, ("xt", "yt", "zt"))[1:-2, 2:-2]
    for kr in range(2):
        ki = 0 if kr == 1 else 1
        for ip in range(2):
            drodxe = (
                drdT[1 + ip : -2 + ip, 2:-2, ki:] * dTdx[1:-2, 2:-2, ki:]
                + drdS[1 + ip : -2 + ip, 2:-2, ki:] * dSdx[1:-2, 2:-2, ki:]
            )
            drodze = (
                drdT[1 + ip : -2 + ip, 2:-2, ki:] * dTdz[1 + ip : -2 + ip, 2:-2, : -1 + kr or None]
                + drdS[1 + ip : -2 + ip, 2:-2, ki:] * dSdz[1 + ip : -2 + ip, 2:-2, : -1 + kr or None]
            )
            sxe = -drodxe / (npx.minimum(0.0, drodze) - epsln)
            taper = dm_taper(sxe, settings.iso_slopec, settings.iso_dslope)
            sumz = update_add(
                sumz,
                at[:, :, ki:],
                vs.dzw[npx.newaxis, npx.newaxis, : -1 + kr or None]
                * vs.maskU[1:-2, 2:-2, ki:]
                * npx.maximum(settings.K_iso_steep, diffloc[1:-2, 2:-2, ki:] * taper),
            )
            vs.Ai_ez = update(vs.Ai_ez, at[1:-2, 2:-2, ki:, ip, kr], taper * sxe * vs.maskU[1:-2, 2:-2, ki:])

    vs.K_11 = update(vs.K_11, at[1:-2, 2:-2, :], sumz / (4.0 * vs.dzt[npx.newaxis, npx.newaxis, :]))

    """
    Compute Ai_nz and K_22 on center of north face of T cell.
    """
    diffloc = update(diffloc, at[...], 0)
    diffloc = update(
        diffloc,
        at[2:-2, 1:-2, 1:],
        0.25
        * (vs.K_iso[2:-2, 1:-2, 1:] + vs.K_iso[2:-2, 1:-2, :-1] + vs.K_iso[2:-2, 2:-1, 1:] + vs.K_iso[2:-2, 2:-1, :-1]),
    )
    diffloc = update(diffloc, at[2:-2, 1:-2, 0], 0.5 * (vs.K_iso[2:-2, 1:-2, 0] + vs.K_iso[2:-2, 2:-1, 0]))

    sumz = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 1:-2]
    for kr in range(2):
        ki = 0 if kr == 1 else 1
        for jp in range(2):
            drodyn = (
                drdT[2:-2, 1 + jp : -2 + jp, ki:] * dTdy[2:-2, 1:-2, ki:]
                + drdS[2:-2, 1 + jp : -2 + jp, ki:] * dSdy[2:-2, 1:-2, ki:]
            )
            drodzn = (
                drdT[2:-2, 1 + jp : -2 + jp, ki:] * dTdz[2:-2, 1 + jp : -2 + jp, : -1 + kr or None]
                + drdS[2:-2, 1 + jp : -2 + jp, ki:] * dSdz[2:-2, 1 + jp : -2 + jp, : -1 + kr or None]
            )
            syn = -drodyn / (npx.minimum(0.0, drodzn) - epsln)
            taper = dm_taper(syn, settings.iso_slopec, settings.iso_dslope)
            sumz = update_add(
                sumz,
                at[:, :, ki:],
                vs.dzw[npx.newaxis, npx.newaxis, : -1 + kr or None]
                * vs.maskV[2:-2, 1:-2, ki:]
                * npx.maximum(settings.K_iso_steep, diffloc[2:-2, 1:-2, ki:] * taper),
            )
            vs.Ai_nz = update(vs.Ai_nz, at[2:-2, 1:-2, ki:, jp, kr], taper * syn * vs.maskV[2:-2, 1:-2, ki:])
    vs.K_22 = update(vs.K_22, at[2:-2, 1:-2, :], sumz / (4.0 * vs.dzt[npx.newaxis, npx.newaxis, :]))

    """
    compute Ai_bx, Ai_by and K33 on top face of T cell.
    """
    sumx = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2, :-1]
    sumy = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2, :-1]

    for kr in range(2):
        drodzb = (
            drdT[2:-2, 2:-2, kr : -1 + kr or None] * dTdz[2:-2, 2:-2, :-1]
            + drdS[2:-2, 2:-2, kr : -1 + kr or None] * dSdz[2:-2, 2:-2, :-1]
        )

        # eastward slopes at the top of T cells
        for ip in range(2):
            drodxb = (
                drdT[2:-2, 2:-2, kr : -1 + kr or None] * dTdx[1 + ip : -3 + ip, 2:-2, kr : -1 + kr or None]
                + drdS[2:-2, 2:-2, kr : -1 + kr or None] * dSdx[1 + ip : -3 + ip, 2:-2, kr : -1 + kr or None]
            )
            sxb = -drodxb / (npx.minimum(0.0, drodzb) - epsln)
            taper = dm_taper(sxb, settings.iso_slopec, settings.iso_dslope)
            sumx = (
                sumx
                + vs.dxu[1 + ip : -3 + ip, npx.newaxis, npx.newaxis]
                * vs.K_iso[2:-2, 2:-2, :-1]
                * taper
                * sxb**2
                * vs.maskW[2:-2, 2:-2, :-1]
            )
            vs.Ai_bx = update(vs.Ai_bx, at[2:-2, 2:-2, :-1, ip, kr], taper * sxb * vs.maskW[2:-2, 2:-2, :-1])

        # northward slopes at the top of T cells
        for jp in range(2):
            facty = vs.cosu[1 + jp : -3 + jp] * vs.dyu[1 + jp : -3 + jp]
            drodyb = (
                drdT[2:-2, 2:-2, kr : -1 + kr or None] * dTdy[2:-2, 1 + jp : -3 + jp, kr : -1 + kr or None]
                + drdS[2:-2, 2:-2, kr : -1 + kr or None] * dSdy[2:-2, 1 + jp : -3 + jp, kr : -1 + kr or None]
            )
            syb = -drodyb / (npx.minimum(0.0, drodzb) - epsln)
            taper = dm_taper(syb, settings.iso_slopec, settings.iso_dslope)
            sumy = (
                sumy
                + facty[npx.newaxis, :, npx.newaxis]
                * vs.K_iso[2:-2, 2:-2, :-1]
                * taper
                * syb**2
                * vs.maskW[2:-2, 2:-2, :-1]
            )
            vs.Ai_by = update(vs.Ai_by, at[2:-2, 2:-2, :-1, jp, kr], taper * syb * vs.maskW[2:-2, 2:-2, :-1])

    vs.K_33 = update(
        vs.K_33,
        at[2:-2, 2:-2, :-1],
        sumx / (4 * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
        + sumy / (4 * vs.dyt[npx.newaxis, 2:-2, npx.newaxis] * vs.cost[npx.newaxis, 2:-2, npx.newaxis]),
    )
    vs.K_33 = update(vs.K_33, at[..., -1], 0.0)

    return KernelOutput(
        Ai_ez=vs.Ai_ez, Ai_nz=vs.Ai_nz, Ai_bx=vs.Ai_bx, Ai_by=vs.Ai_by, K_11=vs.K_11, K_22=vs.K_22, K_33=vs.K_33
    )


@veros_kernel
def isoneutral_diag_streamfunction_kernel(state):
    vs = state.variables

    K_gm_pad = utilities.pad_z_edges(vs.K_gm)

    """
    meridional component at east face of 'T' cells
    """
    diffloc = 0.25 * (
        K_gm_pad[1:-2, 2:-2, 1:-1] + K_gm_pad[1:-2, 2:-2, :-2] + K_gm_pad[2:-1, 2:-2, 1:-1] + K_gm_pad[2:-1, 2:-2, :-2]
    )
    vs.B2_gm = update(vs.B2_gm, at[1:-2, 2:-2, :], 0.25 * diffloc * npx.sum(vs.Ai_ez[1:-2, 2:-2, ...], axis=(3, 4)))

    """
    zonal component at north face of 'T' cells
    """
    diffloc = 0.25 * (
        K_gm_pad[2:-2, 1:-2, 1:-1] + K_gm_pad[2:-2, 1:-2, :-2] + K_gm_pad[2:-2, 2:-1, 1:-1] + K_gm_pad[2:-2, 2:-1, :-2]
    )
    vs.B1_gm = update(vs.B1_gm, at[2:-2, 1:-2, :], -0.25 * diffloc * npx.sum(vs.Ai_nz[2:-2, 1:-2, ...], axis=(3, 4)))

    return KernelOutput(B1_gm=vs.B1_gm, B2_gm=vs.B2_gm)


@veros_routine
def isoneutral_diag_streamfunction(state):
    """
    calculate hor. components of streamfunction for eddy driven velocity
    for diagnostics purpose only
    """
    vs = state.variables
    settings = state.settings

    if not (settings.enable_neutral_diffusion and settings.enable_skew_diffusion):
        return

    vs.update(isoneutral_diag_streamfunction_kernel(state))


@veros_routine
def check_isoneutral_slope_crit(state):
    """
    check linear stability criterion from Griffies et al
    """
    vs = state.variables
    settings = state.settings

    epsln = 1e-20
    if settings.enable_neutral_diffusion:
        ft1 = 1.0 / (4.0 * settings.K_iso_0 * settings.dt_tracer + epsln)
        delta1a = npx.min(
            vs.dxt[2:-2, npx.newaxis, npx.newaxis]
            * npx.abs(vs.cost[npx.newaxis, 2:-2, npx.newaxis])
            * vs.dzt[npx.newaxis, npx.newaxis, :]
            * ft1
        )
        delta1b = npx.min(vs.dyt[npx.newaxis, 2:-2, npx.newaxis] * vs.dzt[npx.newaxis, npx.newaxis, :] * ft1)
        delta_iso1 = min(vs.dzt[0] * ft1 * vs.dxt[-1] * abs(vs.cost[-1]), min(delta1a, delta1b))

        logger.info("Diffusion grid factor delta_iso1 = {}", float(delta_iso1))
        if delta_iso1 < settings.iso_slopec:
            raise RuntimeError(
                "Without latitudinal filtering, delta_iso1 is the steepest "
                "isoneutral slope available for linear stability of "
                "Redi and GM. Maximum allowable isoneutral slope is "
                f"specified as iso_slopec = {settings.iso_slopec}."
            )



#!/usr/bin/env python

import os
import h5netcdf
import scipy.ndimage

from veros import veros_routine, veros_kernel, KernelOutput, VerosSetup, runtime_settings as rs, runtime_state as rst
from veros.variables import Variable, allocate
from veros.core.utilities import enforce_boundaries
from veros.core.operators import numpy as npx, update, at
import veros.tools
import veros.time

BASE_PATH = os.path.dirname(os.path.realpath(__file__))
DATA_FILES = veros.tools.get_assets("global_flexible", os.path.join(BASE_PATH, "assets.json"))


class GlobalFlexibleResolutionSetup(VerosSetup):
    """
    Global model with flexible resolution.
    """

    # global settings
    min_depth = 10.0
    max_depth = 5400.0
    equatorial_grid_spacing_factor = 0.5
    polar_grid_spacing_factor = None

    @veros_routine
    def set_parameter(self, state):
        settings = state.settings

        settings.identifier = "global_flexible"
        settings.description = "Global model with flexible resolution"

        settings.nx = 360
        settings.ny = 160
        settings.nz = 60
        settings.dt_mom = settings.dt_tracer = 900
        settings.runlen = 86400 * 10

        settings.x_origin = 90.0
        settings.y_origin = -80.0

        settings.coord_degree = True
        settings.enable_cyclic_x = True

        # friction
        settings.enable_hor_friction = True
        settings.A_h = 5e4
        settings.enable_hor_friction_cos_scaling = True
        settings.hor_friction_cosPower = 1
        settings.enable_tempsalt_sources = True
        settings.enable_implicit_vert_friction = True

        settings.eq_of_state_type = 5

        # isoneutral
        settings.enable_neutral_diffusion = True
        settings.K_iso_0 = 1000.0
        settings.K_iso_steep = 50.0
        settings.iso_dslope = 0.005
        settings.iso_slopec = 0.005
        settings.enable_skew_diffusion = True

        # tke
        settings.enable_tke = True
        settings.c_k = 0.1
        settings.c_eps = 0.7
        settings.alpha_tke = 30.0
        settings.mxl_min = 1e-8
        settings.tke_mxl_choice = 2
        settings.kappaM_min = 2e-4
        settings.kappaH_min = 2e-5
        settings.enable_kappaH_profile = True
        settings.enable_tke_superbee_advection = True

        # eke
        settings.enable_eke = True
        settings.eke_k_max = 1e4
        settings.eke_c_k = 0.4
        settings.eke_c_eps = 0.5
        settings.eke_cross = 2.0
        settings.eke_crhin = 1.0
        settings.eke_lmin = 100.0
        settings.enable_eke_superbee_advection = True
        settings.enable_eke_isopycnal_diffusion = True

        # idemix
        settings.enable_idemix = False
        settings.enable_eke_diss_surfbot = True
        settings.eke_diss_surfbot_frac = 0.2
        settings.enable_idemix_superbee_advection = True
        settings.enable_idemix_hor_diffusion = True

        # custom variables
        state.dimensions["nmonths"] = 12
        state.var_meta.update(
            t_star=Variable("t_star", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            s_star=Variable("s_star", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            qnec=Variable("qnec", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            qnet=Variable("qnet", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            qsol=Variable("qsol", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            divpen_shortwave=Variable("divpen_shortwave", ("zt",), "", "", time_dependent=False),
            taux=Variable("taux", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
            tauy=Variable("tauy", ("xt", "yt", "nmonths"), "", "", time_dependent=False),
        )

    def _get_data(self, var, idx=None):
        if idx is None:
            idx = Ellipsis
        else:
            idx = idx[::-1]

        kwargs = {}
        if rst.proc_num > 1:
            kwargs.update(
                driver="mpio",
                comm=rs.mpi_comm,
            )

        with h5netcdf.File(DATA_FILES["forcing"], "r", **kwargs) as forcing_file:
            var_obj = forcing_file.variables[var]
            return npx.array(var_obj[idx]).T

    @veros_routine(dist_safe=False, local_variables=["dxt", "dyt", "dzt"])
    def set_grid(self, state):
        vs = state.variables
        settings = state.settings

        if settings.ny % 2:
            raise ValueError("ny has to be an even number of grid cells")

        vs.dxt = update(vs.dxt, at[...], 360.0 / settings.nx)

        if self.equatorial_grid_spacing_factor is not None:
            eq_spacing = self.equatorial_grid_spacing_factor * 160.0 / settings.ny
        else:
            eq_spacing = None

        if self.polar_grid_spacing_factor is not None:
            polar_spacing = self.polar_grid_spacing_factor * 160.0 / settings.ny
        else:
            polar_spacing = None

        vs.dyt = update(
            vs.dyt,
            at[2:-2],
            veros.tools.get_vinokur_grid_steps(
                settings.ny, 160.0, eq_spacing, upper_stepsize=polar_spacing, two_sided_grid=True
            ),
        )
        vs.dzt = veros.tools.get_vinokur_grid_steps(settings.nz, self.max_depth, self.min_depth, refine_towards="lower")

    @veros_routine
    def set_coriolis(self, state):
        vs = state.variables
        settings = state.settings
        vs.coriolis_t = update(
            vs.coriolis_t, at[...], 2 * settings.omega * npx.sin(vs.yt[npx.newaxis, :] / 180.0 * settings.pi)
        )

    def _shift_longitude_array(self, vs, lon, arr):
        wrap_i = npx.where((lon[:-1] < vs.xt.min()) & (lon[1:] >= vs.xt.min()))[0][0]
        new_lon = npx.concatenate((lon[wrap_i:-1], lon[:wrap_i] + 360.0))
        new_arr = npx.concatenate((arr[wrap_i:-1, ...], arr[:wrap_i, ...]))
        return new_lon, new_arr

    @veros_routine(dist_safe=False, local_variables=["kbot", "xt", "yt", "zt"])
    def set_topography(self, state):
        vs = state.variables
        settings = state.settings

        with h5netcdf.File(DATA_FILES["topography"], "r") as topography_file:
            topo_x, topo_y, topo_z = (npx.array(topography_file.variables[k], dtype="float").T for k in ("x", "y", "z"))

        topo_z = npx.minimum(topo_z, 0.0)

        # smooth topography to match grid resolution
        gaussian_sigma = (0.5 * len(topo_x) / settings.nx, 0.5 * len(topo_y) / settings.ny)
        topo_z_smoothed = scipy.ndimage.gaussian_filter(topo_z, sigma=gaussian_sigma)
        topo_z_smoothed = npx.where(topo_z >= -1, 0, topo_z_smoothed)

        topo_x_shifted, topo_z_shifted = self._shift_longitude_array(vs, topo_x, topo_z_smoothed)
        coords = (vs.xt[2:-2], vs.yt[2:-2])
        z_interp = allocate(state.dimensions, ("xt", "yt"), local=False)
        z_interp = update(
            z_interp,
            at[2:-2, 2:-2],
            veros.tools.interpolate((topo_x_shifted, topo_y), topo_z_shifted, coords, kind="nearest", fill=False),
        )

        depth_levels = 1 + npx.argmin(npx.abs(z_interp[:, :, npx.newaxis] - vs.zt[npx.newaxis, npx.newaxis, :]), axis=2)
        vs.kbot = update(vs.kbot, at[2:-2, 2:-2], npx.where(z_interp < 0.0, depth_levels, 0)[2:-2, 2:-2])
        vs.kbot = npx.where(vs.kbot < settings.nz, vs.kbot, 0)
        vs.kbot = enforce_boundaries(vs.kbot, settings.enable_cyclic_x, local=True)

        # remove marginal seas
        # (dilate to close 1-cell passages, fill holes, undo dilation)
        marginal = scipy.ndimage.binary_erosion(
            scipy.ndimage.binary_fill_holes(scipy.ndimage.binary_dilation(vs.kbot == 0))
        )

        vs.kbot = npx.where(marginal, 0, vs.kbot)

    @veros_routine
    def set_initial_conditions(self, state):
        vs = state.variables
        settings = state.settings

        rpart_shortwave = 0.58
        efold1_shortwave = 0.35
        efold2_shortwave = 23.0

        t_grid = (vs.xt[2:-2], vs.yt[2:-2], vs.zt)
        xt_forc, yt_forc, zt_forc = (self._get_data(k) for k in ("xt", "yt", "zt"))
        zt_forc = zt_forc[::-1]

        # coordinates must be monotonous for this to work
        assert npx.diff(xt_forc).all() > 0
        assert npx.diff(yt_forc).all() > 0

        # determine slice to read from forcing file
        data_subset = (
            slice(
                max(0, int(npx.argmax(xt_forc >= vs.xt.min())) - 1),
                len(xt_forc) - max(0, int(npx.argmax(xt_forc[::-1] <= vs.xt.max())) - 1),
            ),
            slice(
                max(0, int(npx.argmax(yt_forc >= vs.yt.min())) - 1),
                len(yt_forc) - max(0, int(npx.argmax(yt_forc[::-1] <= vs.yt.max())) - 1),
            ),
            Ellipsis,
        )

        xt_forc = xt_forc[data_subset[0]]
        yt_forc = yt_forc[data_subset[1]]

        # initial conditions
        temp_raw = self._get_data("temperature", idx=data_subset)[..., ::-1]
        temp_data = veros.tools.interpolate((xt_forc, yt_forc, zt_forc), temp_raw, t_grid)
        vs.temp = update(vs.temp, at[2:-2, 2:-2, :, :], (temp_data * vs.maskT[2:-2, 2:-2, :])[..., npx.newaxis])

        salt_raw = self._get_data("salinity", idx=data_subset)[..., ::-1]
        salt_data = veros.tools.interpolate((xt_forc, yt_forc, zt_forc), salt_raw, t_grid)
        vs.salt = update(vs.salt, at[2:-2, 2:-2, :, :], (salt_data * vs.maskT[2:-2, 2:-2, :])[..., npx.newaxis])

        # wind stress on MIT grid
        time_grid = (vs.xt[2:-2], vs.yt[2:-2], npx.arange(12))
        taux_raw = self._get_data("tau_x", idx=data_subset)
        taux_data = veros.tools.interpolate((xt_forc, yt_forc, npx.arange(12)), taux_raw, time_grid)
        vs.taux = update(vs.taux, at[2:-2, 2:-2, :], taux_data)

        tauy_raw = self._get_data("tau_y", idx=data_subset)
        tauy_data = veros.tools.interpolate((xt_forc, yt_forc, npx.arange(12)), tauy_raw, time_grid)
        vs.tauy = update(vs.tauy, at[2:-2, 2:-2, :], tauy_data)

        vs.taux = enforce_boundaries(vs.taux, settings.enable_cyclic_x)
        vs.tauy = enforce_boundaries(vs.tauy, settings.enable_cyclic_x)

        # Qnet and dQ/dT and Qsol
        qnet_raw = self._get_data("q_net", idx=data_subset)
        qnet_data = veros.tools.interpolate((xt_forc, yt_forc, npx.arange(12)), qnet_raw, time_grid)
        vs.qnet = update(vs.qnet, at[2:-2, 2:-2, :], -qnet_data * vs.maskT[2:-2, 2:-2, -1, npx.newaxis])

        qnec_raw = self._get_data("dqdt", idx=data_subset)
        qnec_data = veros.tools.interpolate((xt_forc, yt_forc, npx.arange(12)), qnec_raw, time_grid)
        vs.qnec = update(vs.qnec, at[2:-2, 2:-2, :], qnec_data * vs.maskT[2:-2, 2:-2, -1, npx.newaxis])

        qsol_raw = self._get_data("swf", idx=data_subset)
        qsol_data = veros.tools.interpolate((xt_forc, yt_forc, npx.arange(12)), qsol_raw, time_grid)
        vs.qsol = update(vs.qsol, at[2:-2, 2:-2, :], -qsol_data * vs.maskT[2:-2, 2:-2, -1, npx.newaxis])

        # SST and SSS
        sst_raw = self._get_data("sst", idx=data_subset)
        sst_data = veros.tools.interpolate((xt_forc, yt_forc, npx.arange(12)), sst_raw, time_grid)
        vs.t_star = update(vs.t_star, at[2:-2, 2:-2, :], sst_data * vs.maskT[2:-2, 2:-2, -1, npx.newaxis])

        sss_raw = self._get_data("sss", idx=data_subset)
        sss_data = veros.tools.interpolate((xt_forc, yt_forc, npx.arange(12)), sss_raw, time_grid)
        vs.s_star = update(vs.s_star, at[2:-2, 2:-2, :], sss_data * vs.maskT[2:-2, 2:-2, -1, npx.newaxis])

        if settings.enable_idemix:
            tidal_energy_raw = self._get_data("tidal_energy", idx=data_subset)
            tidal_energy_data = veros.tools.interpolate((xt_forc, yt_forc), tidal_energy_raw, t_grid[:-1])
            mask_x, mask_y = (i + 2 for i in npx.indices((vs.nx, vs.ny)))
            mask_z = npx.maximum(0, vs.kbot[2:-2, 2:-2] - 1)
            tidal_energy_data[:, :] *= vs.maskW[mask_x, mask_y, mask_z] / vs.rho_0
            vs.forc_iw_bottom[2:-2, 2:-2] = tidal_energy_data

        """
        Initialize penetration profile for solar radiation and store divergence in divpen
        note that pen is set to 0.0 at the surface instead of 1.0 to compensate for the
        shortwave part of the total surface flux
        """
        swarg1 = vs.zw / efold1_shortwave
        swarg2 = vs.zw / efold2_shortwave
        pen = rpart_shortwave * npx.exp(swarg1) + (1.0 - rpart_shortwave) * npx.exp(swarg2)
        pen = update(pen, at[-1], 0.0)
        vs.divpen_shortwave = update(vs.divpen_shortwave, at[1:], (pen[1:] - pen[:-1]) / vs.dzt[1:])
        vs.divpen_shortwave = update(vs.divpen_shortwave, at[0], pen[0] / vs.dzt[0])

    @veros_routine
    def set_forcing(self, state):
        vs = state.variables
        vs.update(set_forcing_kernel(state))

    @veros_routine
    def set_diagnostics(self, state):
        settings = state.settings
        diagnostics = state.diagnostics

        diagnostics["cfl_monitor"].output_frequency = settings.dt_tracer * 100
        diagnostics["tracer_monitor"].output_frequency = settings.dt_tracer * 100
        diagnostics["snapshot"].output_frequency = 30 * 86400.0
        diagnostics["overturning"].output_frequency = 360 * 86400
        diagnostics["overturning"].sampling_frequency = 86400.0
        diagnostics["energy"].output_frequency = 360 * 86400
        diagnostics["energy"].sampling_frequency = 10 * settings.dt_tracer
        diagnostics["averages"].output_frequency = 30 * 86400
        diagnostics["averages"].sampling_frequency = settings.dt_tracer

        average_vars = [
            "surface_taux",
            "surface_tauy",
            "forc_temp_surface",
            "forc_salt_surface",
            "psi",
            "temp",
            "salt",
            "u",
            "v",
            "w",
            "Nsqr",
            "Hd",
            "rho",
            "kappaH",
        ]
        if settings.enable_skew_diffusion:
            average_vars += ["B1_gm", "B2_gm"]
        if settings.enable_TEM_friction:
            average_vars += ["kappa_gm", "K_diss_gm"]
        if settings.enable_tke:
            average_vars += ["tke", "Prandtlnumber", "mxl", "tke_diss", "forc_tke_surface", "tke_surf_corr"]
        if settings.enable_idemix:
            average_vars += ["E_iw", "forc_iw_surface", "iw_diss", "c0", "v0"]
        if settings.enable_eke:
            average_vars += ["eke", "K_gm", "L_rossby", "L_rhines"]
        diagnostics["averages"].output_variables = average_vars

    @veros_routine
    def after_timestep(self, state):
        pass


@veros_kernel
def set_forcing_kernel(state):
    vs = state.variables
    settings = state.settings

    t_rest = 30.0 * 86400.0
    cp_0 = 3991.86795711963  # J/kg /K

    year_in_seconds = veros.time.convert_time(1.0, "years", "seconds")
    (n1, f1), (n2, f2) = veros.tools.get_periodic_interval(vs.time, year_in_seconds, year_in_seconds / 12.0, 12)

    # linearly interpolate wind stress and shift from MITgcm U/V grid to this grid
    vs.surface_taux = update(vs.surface_taux, at[:-1, :], f1 * vs.taux[1:, :, n1] + f2 * vs.taux[1:, :, n2])
    vs.surface_tauy = update(vs.surface_tauy, at[:, :-1], f1 * vs.tauy[:, 1:, n1] + f2 * vs.tauy[:, 1:, n2])

    if settings.enable_tke:
        vs.forc_tke_surface = update(
            vs.forc_tke_surface,
            at[1:-1, 1:-1],
            npx.sqrt(
                (0.5 * (vs.surface_taux[1:-1, 1:-1] + vs.surface_taux[:-2, 1:-1]) / settings.rho_0) ** 2
                + (0.5 * (vs.surface_tauy[1:-1, 1:-1] + vs.surface_tauy[1:-1, :-2]) / settings.rho_0) ** 2
            )
            ** (3.0 / 2.0),
        )

    # W/m^2 K kg/J m^3/kg = K m/s
    t_star_cur = f1 * vs.t_star[..., n1] + f2 * vs.t_star[..., n2]
    qqnec = f1 * vs.qnec[..., n1] + f2 * vs.qnec[..., n2]
    qqnet = f1 * vs.qnet[..., n1] + f2 * vs.qnet[..., n2]
    vs.forc_temp_surface = (
        (qqnet + qqnec * (t_star_cur - vs.temp[..., -1, vs.tau])) * vs.maskT[..., -1] / cp_0 / settings.rho_0
    )
    s_star_cur = f1 * vs.s_star[..., n1] + f2 * vs.s_star[..., n2]
    vs.forc_salt_surface = 1.0 / t_rest * (s_star_cur - vs.salt[..., -1, vs.tau]) * vs.maskT[..., -1] * vs.dzt[-1]

    # apply simple ice mask
    mask1 = vs.temp[:, :, -1, vs.tau] * vs.maskT[:, :, -1] > -1.8
    mask2 = vs.forc_temp_surface > 0
    ice = npx.logical_or(mask1, mask2)
    vs.forc_temp_surface *= ice
    vs.forc_salt_surface *= ice

    # solar radiation
    if settings.enable_tempsalt_sources:
        vs.temp_source = (
            (f1 * vs.qsol[..., n1, None] + f2 * vs.qsol[..., n2, None])
            * vs.divpen_shortwave[None, None, :]
            * ice[..., None]
            * vs.maskT[..., :]
            / cp_0
            / settings.rho_0
        )

    return KernelOutput(
        surface_taux=vs.surface_taux,
        surface_tauy=vs.surface_tauy,
        temp_source=vs.temp_source,
        forc_tke_surface=vs.forc_tke_surface,
        forc_temp_surface=vs.forc_temp_surface,
        forc_salt_surface=vs.forc_salt_surface,
    )



# defensive imports since extensions are optional
try:
    from veros.core.special import tdma_cython_
except ImportError:
    HAS_CPU_EXT = False
else:
    HAS_CPU_EXT = True

try:
    from veros.core.special import tdma_cuda_
except ImportError:
    HAS_GPU_EXT = False
else:
    HAS_GPU_EXT = True

import numpy as np

import jax
import jax.numpy as jnp
from jax.extend.core import Primitive
from jax.core import ShapedArray
from jax.interpreters import xla, mlir
import jaxlib.mlir.ir as ir
from jaxlib.mlir.dialects import mhlo

from jax._src.interpreters.mlir import custom_call


if HAS_CPU_EXT:
    for kernel_name in ("tdma_cython_double", "tdma_cython_float"):
        fn = tdma_cython_.cpu_custom_call_targets[kernel_name]
        jax.ffi.register_ffi_target(kernel_name, fn, platform="cpu", api_version=0)

if HAS_GPU_EXT:
    for kernel_name in ("tdma_cuda_double", "tdma_cuda_float"):
        fn = tdma_cuda_.gpu_custom_call_targets[kernel_name]
        jax.ffi.register_ffi_target(kernel_name, fn, platform="CUDA", api_version=0)


def as_mhlo_constant(val, dtype):
    if isinstance(val, mhlo.ConstantOp):
        return val

    return mhlo.ConstantOp(
        ir.DenseElementsAttr.get(np.array([val], dtype=dtype), type=mlir.dtype_to_ir_type(np.dtype(dtype)))
    ).result


def tdma(a, b, c, d, interior_mask, edge_mask, device=None):
    if device is None:
        device = jax.default_backend()

    if not a.shape == b.shape == c.shape == d.shape:
        raise ValueError("all inputs must have identical shape")

    if not a.dtype == b.dtype == c.dtype == d.dtype:
        raise ValueError("all inputs must have the same dtype")

    if device == "cpu":
        system_depths = jnp.sum(interior_mask, axis=-1, dtype="int32")
        return tdma_p.bind(a, b, c, d, system_depths)

    a = interior_mask * a * jnp.logical_not(edge_mask)
    b = jnp.where(interior_mask, b, 1.0)
    c = interior_mask * c
    d = interior_mask * d

    return tdma_p.bind(a, b, c, d, system_depths=None)


def tdma_impl(*args, **kwargs):
    return xla.apply_primitive(tdma_p, *args, **kwargs)


def tdma_xla_encode_cpu(ctx, a, b, c, d, system_depths):
    # try import again to trigger exception on ImportError
    from veros.core.special import tdma_cython_  # noqa: F401

    x_aval, *_ = ctx.avals_in
    np_dtype = x_aval.dtype

    x_type = ir.RankedTensorType(a.type)
    dtype = x_type.element_type
    dims = x_type.shape

    supported_dtypes = (
        np.dtype(np.float32),
        np.dtype(np.float64),
    )

    if np_dtype not in supported_dtypes:
        raise TypeError(f"TDMA only supports {supported_dtypes} arrays, got: {dtype}")

    # compute number of elements to vectorize over
    num_systems = 1
    for el in dims[:-1]:
        num_systems *= el

    stride = dims[-1]

    out_types = [
        ir.RankedTensorType.get(dims, dtype),
        ir.RankedTensorType.get((stride,), dtype),
    ]

    if np_dtype is np.dtype(np.float32):
        kernel = b"tdma_cython_float"
    elif np_dtype is np.dtype(np.float64):
        kernel = b"tdma_cython_double"
    else:
        raise RuntimeError("got unrecognized dtype")

    out = custom_call(
        kernel,
        operands=(
            a,
            b,
            c,
            d,
            system_depths,
            as_mhlo_constant(num_systems, np.int64),
            as_mhlo_constant(stride, np.int64),
        ),
        result_types=out_types,
    )
    return out.results[:-1]


def tdma_xla_encode_gpu(ctx, a, b, c, d, system_depths):
    # try import again to trigger exception on ImportError
    from veros.core.special import tdma_cuda_  # noqa: F401

    if system_depths is not None:
        raise ValueError("TDMA does not support system_depths argument on GPU")

    x_aval, *_ = ctx.avals_in
    np_dtype = x_aval.dtype

    x_type = ir.RankedTensorType(a.type)
    dtype = x_type.element_type
    dims = x_type.shape

    supported_dtypes = (
        np.dtype(np.float32),
        np.dtype(np.float64),
    )

    if np_dtype not in supported_dtypes:
        raise TypeError(f"TDMA only supports {supported_dtypes} arrays, got: {dtype}")

    # compute number of elements to vectorize over
    num_systems = 1
    for el in dims[:-1]:
        num_systems *= el

    system_depth = dims[-1]

    if np_dtype is np.dtype(np.float32):
        kernel = b"tdma_cuda_float"
    elif np_dtype is np.dtype(np.float64):
        kernel = b"tdma_cuda_double"
    else:
        raise RuntimeError("got unrecognized dtype")

    descriptor = tdma_cuda_.build_tridiag_descriptor(num_systems, system_depth)

    ndims = len(dims)
    arr_layout = tuple(range(ndims - 2, -1, -1)) + (ndims - 1,)

    out_types = [ir.RankedTensorType.get(dims, dtype), ir.RankedTensorType.get(dims, dtype)]
    out_layouts = (arr_layout, arr_layout)

    out = custom_call(
        kernel,
        operands=(a, b, c, d),
        result_types=out_types,
        result_layouts=out_layouts,
        operand_layouts=(arr_layout,) * 4,
        backend_config=descriptor,
    )
    return out.results[:-1]


def tdma_abstract_eval(a, b, c, d, system_depths):
    return ShapedArray(a.shape, a.dtype)


tdma_p = Primitive("tdma")
tdma_p.def_impl(tdma_impl)
tdma_p.def_abstract_eval(tdma_abstract_eval)

mlir.register_lowering(tdma_p, tdma_xla_encode_cpu, platform="cpu")
mlir.register_lowering(tdma_p, tdma_xla_encode_gpu, platform="cuda")



from veros.core.operators import update, at, numpy as npx
from veros.variables import allocate


def assemble_poisson_matrix(state):
    if state.settings.enable_streamfunction:
        return assemble_streamfunction_matrix(state)
    else:
        return assemble_pressure_matrix(state)


def assemble_pressure_matrix(state):
    main_diag = allocate(state.dimensions, ("xu", "yu"), fill=1)
    east_diag, west_diag, north_diag, south_diag = (allocate(state.dimensions, ("xu", "yu")) for _ in range(4))

    vs = state.variables
    settings = state.settings

    maskM = vs.maskT[:, :, -1]

    mp_i = maskM[2:-2, 2:-2] * maskM[3:-1, 2:-2]
    mm_i = maskM[2:-2, 2:-2] * maskM[1:-3, 2:-2]

    mp_j = maskM[2:-2, 2:-2] * maskM[2:-2, 3:-1]
    mm_j = maskM[2:-2, 2:-2] * maskM[2:-2, 1:-3]

    main_diag = update(
        main_diag,
        at[2:-2, 2:-2],
        -1
        * mp_i
        * vs.hu[2:-2, 2:-2]
        / vs.dxu[2:-2, npx.newaxis]
        / vs.dxt[2:-2, npx.newaxis]
        / vs.cost[npx.newaxis, 2:-2] ** 2
        - 1
        * mm_i
        * vs.hu[1:-3, 2:-2]
        / vs.dxu[1:-3, npx.newaxis]
        / vs.dxt[2:-2, npx.newaxis]
        / vs.cost[npx.newaxis, 2:-2] ** 2
        - 1
        * mp_j
        * vs.hv[2:-2, 2:-2]
        / vs.dyu[npx.newaxis, 2:-2]
        / vs.dyt[npx.newaxis, 2:-2]
        * vs.cosu[npx.newaxis, 2:-2]
        / vs.cost[npx.newaxis, 2:-2]
        - 1
        * mm_j
        * vs.hv[2:-2, 1:-3]
        / vs.dyu[npx.newaxis, 1:-3]
        / vs.dyt[npx.newaxis, 2:-2]
        * vs.cosu[npx.newaxis, 1:-3]
        / vs.cost[npx.newaxis, 2:-2]
        # free surface
        - 1.0 / (settings.grav * settings.dt_mom * settings.dt_tracer) * maskM[2:-2, 2:-2],
    )

    east_diag = update(
        east_diag,
        at[2:-2, 2:-2],
        mp_i
        * vs.hu[2:-2, 2:-2]
        / vs.dxu[2:-2, npx.newaxis]
        / vs.dxt[2:-2, npx.newaxis]
        / vs.cost[npx.newaxis, 2:-2] ** 2,
    )

    west_diag = update(
        west_diag,
        at[2:-2, 2:-2],
        mm_i
        * vs.hu[1:-3, 2:-2]
        / vs.dxu[1:-3, npx.newaxis]
        / vs.dxt[2:-2, npx.newaxis]
        / vs.cost[npx.newaxis, 2:-2] ** 2,
    )

    north_diag = update(
        north_diag,
        at[2:-2, 2:-2],
        mp_j
        * vs.hv[2:-2, 2:-2]
        / vs.dyu[npx.newaxis, 2:-2]
        / vs.dyt[npx.newaxis, 2:-2]
        * vs.cosu[npx.newaxis, 2:-2]
        / vs.cost[npx.newaxis, 2:-2],
    )

    south_diag = update(
        south_diag,
        at[2:-2, 2:-2],
        mm_j
        * vs.hv[2:-2, 1:-3]
        / vs.dyu[npx.newaxis, 1:-3]
        / vs.dyt[npx.newaxis, 2:-2]
        * vs.cosu[npx.newaxis, 1:-3]
        / vs.cost[npx.newaxis, 2:-2],
    )
    main_diag = main_diag * maskM
    main_diag = npx.where(npx.abs(main_diag) == 0.0, 1, main_diag)

    offsets = [(0, 0), (1, 0), (-1, 0), (0, 1), (0, -1)]
    diags = [main_diag, east_diag, west_diag, north_diag, south_diag]

    return diags, offsets, maskM


def assemble_streamfunction_matrix(state):
    vs = state.variables

    # assemble diagonals
    main_diag = allocate(state.dimensions, ("xu", "yu"), fill=1)
    east_diag, west_diag, north_diag, south_diag = (allocate(state.dimensions, ("xu", "yu")) for _ in range(4))
    main_diag = update(
        main_diag,
        at[2:-2, 2:-2],
        -vs.hvr[3:-1, 2:-2] / vs.dxu[2:-2, npx.newaxis] / vs.dxt[3:-1, npx.newaxis] / vs.cosu[npx.newaxis, 2:-2] ** 2
        - vs.hvr[2:-2, 2:-2] / vs.dxu[2:-2, npx.newaxis] / vs.dxt[2:-2, npx.newaxis] / vs.cosu[npx.newaxis, 2:-2] ** 2
        - vs.hur[2:-2, 2:-2]
        / vs.dyu[npx.newaxis, 2:-2]
        / vs.dyt[npx.newaxis, 2:-2]
        * vs.cost[npx.newaxis, 2:-2]
        / vs.cosu[npx.newaxis, 2:-2]
        - vs.hur[2:-2, 3:-1]
        / vs.dyu[npx.newaxis, 2:-2]
        / vs.dyt[npx.newaxis, 3:-1]
        * vs.cost[npx.newaxis, 3:-1]
        / vs.cosu[npx.newaxis, 2:-2],
    )
    east_diag = update(
        east_diag,
        at[2:-2, 2:-2],
        vs.hvr[3:-1, 2:-2] / vs.dxu[2:-2, npx.newaxis] / vs.dxt[3:-1, npx.newaxis] / vs.cosu[npx.newaxis, 2:-2] ** 2,
    )
    west_diag = update(
        west_diag,
        at[2:-2, 2:-2],
        vs.hvr[2:-2, 2:-2] / vs.dxu[2:-2, npx.newaxis] / vs.dxt[2:-2, npx.newaxis] / vs.cosu[npx.newaxis, 2:-2] ** 2,
    )
    north_diag = update(
        north_diag,
        at[2:-2, 2:-2],
        vs.hur[2:-2, 3:-1]
        / vs.dyu[npx.newaxis, 2:-2]
        / vs.dyt[npx.newaxis, 3:-1]
        * vs.cost[npx.newaxis, 3:-1]
        / vs.cosu[npx.newaxis, 2:-2],
    )
    south_diag = update(
        south_diag,
        at[2:-2, 2:-2],
        vs.hur[2:-2, 2:-2]
        / vs.dyu[npx.newaxis, 2:-2]
        / vs.dyt[npx.newaxis, 2:-2]
        * vs.cost[npx.newaxis, 2:-2]
        / vs.cosu[npx.newaxis, 2:-2],
    )

    main_diag = main_diag * vs.isle_boundary_mask
    main_diag = npx.where(main_diag == 0.0, 1.0, main_diag)

    offsets = [(0, 0), (1, 0), (-1, 0), (0, 1), (0, -1)]
    diags = [
        main_diag,
        east_diag * vs.isle_boundary_mask,
        west_diag * vs.isle_boundary_mask,
        north_diag * vs.isle_boundary_mask,
        south_diag * vs.isle_boundary_mask,
    ]

    return diags, offsets, vs.isle_boundary_mask



#!/usr/bin/env python

from veros import VerosSetup, veros_routine
from veros.variables import allocate, Variable
from veros.distributed import global_min, global_max
from veros.core.operators import numpy as npx, update, at


class ACCBasicSetup(VerosSetup):
    """A model using spherical coordinates with a partially closed domain representing the Atlantic and ACC.

    Wind forcing over the channel part and buoyancy relaxation drive a large-scale meridional overturning circulation.

    This setup demonstrates:
     - setting up an idealized geometry
     - updating surface forcings
     - constant horizontal and vertical mixing (switched off IDEMIX and GM_EKE)
     - basic usage of diagnostics

    :doc:`Adapted from ACC channel model </reference/setups/acc>`.
    """

    @veros_routine
    def set_parameter(self, state):
        settings = state.settings
        settings.identifier = "acc_basic"
        settings.description = "My ACC basic setup"

        settings.nx, settings.ny, settings.nz = 30, 42, 15
        settings.dt_mom = 4800
        settings.dt_tracer = 86400 / 2.0
        settings.runlen = 86400 * 365 * 20

        settings.x_origin = 0.0
        settings.y_origin = -40.0

        settings.coord_degree = True
        settings.enable_cyclic_x = True

        settings.enable_neutral_diffusion = True
        settings.K_iso_0 = 1000.0
        settings.K_iso_steep = 500.0
        settings.iso_dslope = 0.005
        settings.iso_slopec = 0.01
        settings.enable_skew_diffusion = True

        settings.enable_hor_friction = True
        settings.A_h = (2 * settings.degtom) ** 3 * 2e-11
        settings.enable_hor_friction_cos_scaling = True
        settings.hor_friction_cosPower = 1

        settings.enable_bottom_friction = True
        settings.r_bot = 1e-5

        settings.enable_implicit_vert_friction = True

        settings.enable_tke = True
        settings.c_k = 0.1
        settings.c_eps = 0.7
        settings.alpha_tke = 30.0
        settings.mxl_min = 1e-8
        settings.tke_mxl_choice = 2
        settings.kappaM_min = 2e-4
        settings.kappaH_min = 2e-5
        settings.enable_Prandtl_tke = False
        settings.enable_kappaH_profile = True

        settings.K_gm_0 = 1000.0
        settings.enable_eke = False
        settings.enable_idemix = False

        settings.eq_of_state_type = 3

        var_meta = state.var_meta
        var_meta.update(
            t_star=Variable("t_star", ("yt",), "deg C", "Reference surface temperature"),
            t_rest=Variable("t_rest", ("xt", "yt"), "1/s", "Surface temperature restoring time scale"),
        )

    @veros_routine
    def set_grid(self, state):
        vs = state.variables

        ddz = npx.array(
            [50.0, 70.0, 100.0, 140.0, 190.0, 240.0, 290.0, 340.0, 390.0, 440.0, 490.0, 540.0, 590.0, 640.0, 690.0]
        )
        vs.dxt = update(vs.dxt, at[...], 2.0)
        vs.dyt = update(vs.dyt, at[...], 2.0)
        vs.dzt = ddz[::-1] / 2.5

    @veros_routine
    def set_coriolis(self, state):
        vs = state.variables
        settings = state.settings
        vs.coriolis_t = update(
            vs.coriolis_t, at[:, :], 2 * settings.omega * npx.sin(vs.yt[None, :] / 180.0 * settings.pi)
        )

    @veros_routine
    def set_topography(self, state):
        vs = state.variables
        x, y = npx.meshgrid(vs.xt, vs.yt, indexing="ij")
        vs.kbot = npx.logical_or(x > 1.0, y < -20).astype("int")

    @veros_routine
    def set_initial_conditions(self, state):
        vs = state.variables
        settings = state.settings

        # initial conditions
        vs.temp = update(vs.temp, at[...], ((1 - vs.zt[None, None, :] / vs.zw[0]) * 15 * vs.maskT)[..., None])
        vs.salt = update(vs.salt, at[...], 35.0 * vs.maskT[..., None])

        # wind stress forcing
        yt_min = global_min(vs.yt.min())
        yu_min = global_min(vs.yu.min())
        yt_max = global_max(vs.yt.max())
        yu_max = global_max(vs.yu.max())

        taux = allocate(state.dimensions, ("yt",))
        taux = npx.where(vs.yt < -20, 0.1 * npx.sin(settings.pi * (vs.yu - yu_min) / (-20.0 - yt_min)), taux)
        taux = npx.where(vs.yt > 10, 0.1 * (1 - npx.cos(2 * settings.pi * (vs.yu - 10.0) / (yu_max - 10.0))), taux)
        vs.surface_taux = taux * vs.maskU[:, :, -1]

        # surface heatflux forcing
        vs.t_star = allocate(state.dimensions, ("yt",), fill=15)
        vs.t_star = npx.where(vs.yt < -20, 15 * (vs.yt - yt_min) / (-20 - yt_min), vs.t_star)
        vs.t_star = npx.where(vs.yt > 20, 15 * (1 - (vs.yt - 20) / (yt_max - 20)), vs.t_star)
        vs.t_rest = vs.dzt[npx.newaxis, -1] / (30.0 * 86400.0) * vs.maskT[:, :, -1]

        if settings.enable_tke:
            vs.forc_tke_surface = update(
                vs.forc_tke_surface,
                at[2:-2, 2:-2],
                npx.sqrt(
                    (0.5 * (vs.surface_taux[2:-2, 2:-2] + vs.surface_taux[1:-3, 2:-2]) / settings.rho_0) ** 2
                    + (0.5 * (vs.surface_tauy[2:-2, 2:-2] + vs.surface_tauy[2:-2, 1:-3]) / settings.rho_0) ** 2
                )
                ** (1.5),
            )

    @veros_routine
    def set_forcing(self, state):
        vs = state.variables
        vs.forc_temp_surface = vs.t_rest * (vs.t_star - vs.temp[:, :, -1, vs.tau])

    @veros_routine
    def set_diagnostics(self, state):
        settings = state.settings
        diagnostics = state.diagnostics

        diagnostics["averages"].output_variables = (
            "salt",
            "temp",
            "u",
            "v",
            "w",
            "psi",
            "surface_taux",
            "surface_tauy",
        )
        diagnostics["averages"].output_frequency = 365 * 86400.0
        diagnostics["averages"].sampling_frequency = settings.dt_tracer * 10
        diagnostics["overturning"].output_frequency = 365 * 86400.0 / 48.0
        diagnostics["overturning"].sampling_frequency = settings.dt_tracer * 10
        diagnostics["tracer_monitor"].output_frequency = 365 * 86400.0 / 12.0

    @veros_routine
    def after_timestep(self, state):
        pass



"""
solve two dimensional Possion equation
    A * dpsi = forc,  where A = nabla_h^2
with Neumann boundary conditions
used for surface pressure or free surface
method same as pressure method in MITgcm
"""


from veros import veros_routine
from veros.routines import veros_kernel
from veros.state import KernelOutput
from veros.variables import allocate
from veros.core import utilities as mainutils
from veros.core.operators import update, update_add, at, for_loop
from veros.core.operators import numpy as npx
from veros.core.external.solvers import get_linear_solver


@veros_routine
def solve_pressure(state):
    vs = state.variables
    state_update, forc = prepare_forcing(state)
    vs.update(state_update)

    linear_solver = get_linear_solver(state)
    linear_sol = linear_solver.solve(state, forc, vs.psi[..., vs.taup1])
    linear_sol = mainutils.enforce_boundaries(linear_sol, state.settings.enable_cyclic_x)

    if vs.itt == 0:
        vs.psi = update(vs.psi, at[...], linear_sol[..., npx.newaxis])
    else:
        vs.psi = update(vs.psi, at[..., vs.taup1], linear_sol)

    vs.update(barotropic_velocity_update(state))


@veros_kernel
def prepare_forcing(state):
    vs = state.variables
    settings = state.settings

    # hydrostatic pressure
    vs.p_hydro = update(
        vs.p_hydro,
        at[:, :, -1],
        0.5 * vs.rho[:, :, -1, vs.tau] * settings.grav / settings.rho_0 * vs.dzw[-1] * vs.maskT[:, :, -1],
    )

    def compute_p_hydro(k_inv, p_hydro):
        k = settings.nz - k_inv - 1
        p_hydro = update(
            p_hydro,
            at[..., k],
            vs.maskT[:, :, k]
            * (
                p_hydro[:, :, k + 1]
                + 0.5
                * vs.dzw[k]
                * settings.grav
                / settings.rho_0
                * (vs.rho[:, :, k + 1, vs.tau] + vs.rho[:, :, k, vs.tau])
            ),
        )
        return p_hydro

    vs.p_hydro = for_loop(1, settings.nz, compute_p_hydro, vs.p_hydro)

    # add hydrostatic pressure gradient
    vs.du = update_add(
        vs.du,
        at[2:-2, 2:-2, :, vs.tau],
        -(vs.p_hydro[3:-1, 2:-2, :] - vs.p_hydro[2:-2, 2:-2, :])
        / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxu[2:-2, npx.newaxis, npx.newaxis])
        * vs.maskU[2:-2, 2:-2, :],
    )
    vs.dv = update_add(
        vs.dv,
        at[2:-2, 2:-2, :, vs.tau],
        -(vs.p_hydro[2:-2, 3:-1, :] - vs.p_hydro[2:-2, 2:-2, :])
        / vs.dyu[npx.newaxis, 2:-2, npx.newaxis]
        * vs.maskV[2:-2, 2:-2, :],
    )

    # Integrate forward in time
    vs.u = update(
        vs.u,
        at[:, :, :, vs.taup1],
        vs.u[:, :, :, vs.tau]
        + settings.dt_mom
        * (
            vs.du_mix
            + (1.5 + settings.AB_eps) * vs.du[:, :, :, vs.tau]
            - (0.5 + settings.AB_eps) * vs.du[:, :, :, vs.taum1]
        )
        * vs.maskU,
    )

    vs.v = update(
        vs.v,
        at[:, :, :, vs.taup1],
        vs.v[:, :, :, vs.tau]
        + settings.dt_mom
        * (
            vs.dv_mix
            + (1.5 + settings.AB_eps) * vs.dv[:, :, :, vs.tau]
            - (0.5 + settings.AB_eps) * vs.dv[:, :, :, vs.taum1]
        )
        * vs.maskV,
    )

    # forcing for surface pressure
    uloc = allocate(state.dimensions, ("xt", "yt"))
    vloc = allocate(state.dimensions, ("xt", "yt"))

    uloc = update(
        uloc,
        at[2:-2, 2:-2],
        npx.sum((vs.u[2:-2, 2:-2, :, vs.taup1]) * vs.maskU[2:-2, 2:-2, :] * vs.dzt, axis=(2,)) / settings.dt_mom,
    )
    vloc = update(
        vloc,
        at[2:-2, 2:-2],
        npx.sum((vs.v[2:-2, 2:-2, :, vs.taup1]) * vs.maskV[2:-2, 2:-2, :] * vs.dzt, axis=(2,)) / settings.dt_mom,
    )

    uloc = mainutils.enforce_boundaries(uloc, settings.enable_cyclic_x)
    vloc = mainutils.enforce_boundaries(vloc, settings.enable_cyclic_x)

    forc = allocate(state.dimensions, ("xt", "yt"))

    forc = update(
        forc,
        at[2:-2, 2:-2],
        (uloc[2:-2, 2:-2] - uloc[1:-3, 2:-2]) / (vs.cost[2:-2] * vs.dxt[2:-2, npx.newaxis])
        + (vs.cosu[2:-2] * vloc[2:-2, 2:-2] - vs.cosu[1:-3] * vloc[2:-2, 1:-3]) / (vs.cost[2:-2] * vs.dyt[2:-2])
        # free surface
        - vs.psi[2:-2, 2:-2, vs.tau]
        / (settings.grav * settings.dt_mom * settings.dt_tracer)
        * vs.maskT[2:-2, 2:-2, -1],
    )

    # first guess
    vs.psi = update(vs.psi, at[:, :, vs.taup1], 2 * vs.psi[:, :, vs.tau] - vs.psi[:, :, vs.taum1])

    return KernelOutput(du=vs.du, dv=vs.dv, u=vs.u, v=vs.v, psi=vs.psi, p_hydro=vs.p_hydro), forc


@veros_kernel
def barotropic_velocity_update(state):
    """
    solve for surface pressure
    """
    vs = state.variables
    settings = state.settings

    vs.u = update_add(
        vs.u,
        at[2:-2, 2:-2, :, vs.taup1],
        -settings.dt_mom
        * (vs.psi[3:-1, 2:-2, vs.taup1, npx.newaxis] - vs.psi[2:-2, 2:-2, vs.taup1, npx.newaxis])
        / (vs.dxu[2:-2, npx.newaxis, npx.newaxis] * vs.cost[2:-2, npx.newaxis])
        * vs.maskU[2:-2, 2:-2, :],
    )

    vs.v = update_add(
        vs.v,
        at[2:-2, 2:-2, :, vs.taup1],
        -settings.dt_mom
        * (vs.psi[2:-2, 3:-1, vs.taup1, npx.newaxis] - vs.psi[2:-2, 2:-2, vs.taup1, npx.newaxis])
        / vs.dyu[npx.newaxis, 2:-2, npx.newaxis]
        * vs.maskV[2:-2, 2:-2, :],
    )

    vs.ssh = vs.psi[..., vs.tau] / settings.grav

    return KernelOutput(u=vs.u, v=vs.v, ssh=vs.ssh)



import functools
import inspect
import threading
from contextlib import ExitStack, contextmanager

from veros import logger

from veros.state import VerosState


# stack helpers


class RoutineStack:
    def __init__(self):
        self.keep_full_stack = False
        self._stack = []
        self._current_idx = []

    @property
    def stack_level(self):
        return len(self._current_idx)

    def append(self, val):
        frame = self._stack
        for i in self._current_idx:
            frame = frame[i][1]

        self._current_idx.append(len(frame))
        frame.append([val, []])

    def pop(self):
        frame = self._stack
        for i in self._current_idx[:-1]:
            frame = frame[i][1]

        if self.keep_full_stack:
            last_val = frame[-1][0]
        else:
            last_val = frame.pop()[0]
        self._current_idx.pop()
        return last_val


# global context

CURRENT_CONTEXT = threading.local()
CURRENT_CONTEXT.is_dist_safe = True
CURRENT_CONTEXT.routine_stack = RoutineStack()


@contextmanager
def nullcontext():
    yield


@contextmanager
def enter_routine(name, routine_obj, timer=None, dist_safe=True):
    from veros import runtime_state as rst
    from veros.distributed import abort

    stack = CURRENT_CONTEXT.routine_stack

    logger.trace("{}> {}", "-" * stack.stack_level, name)
    stack.append(routine_obj)

    reset_dist_safe = False
    if CURRENT_CONTEXT.is_dist_safe:
        if not dist_safe and rst.proc_num > 1:
            CURRENT_CONTEXT.is_dist_safe = False
            reset_dist_safe = True

    timer_ctx = nullcontext() if timer is None else timer

    try:
        with timer_ctx:
            yield

    except:  # noqa: E722
        if reset_dist_safe:
            abort()
        raise

    finally:
        if reset_dist_safe:
            CURRENT_CONTEXT.is_dist_safe = True

        r = stack.pop()
        assert r is routine_obj

        exec_time = ""
        if timer is not None:
            exec_time = f"({timer.last_time:.3f}s)"

        logger.trace("<{} {} {}", "-" * stack.stack_level, name, exec_time)


# helper functions


def _get_func_name(function):
    return f"{inspect.getmodule(function).__name__}:{function.__qualname__}"


def _is_method(function):
    if inspect.ismethod(function):
        return True

    # hack for unbound methods: check if first argument is called "self"
    spec = inspect.getfullargspec(function)
    return spec.args and spec.args[0] == "self"


# routine


def veros_routine(function=None, *, dist_safe=True, local_variables=()):
    """
    .. note::

      This decorator should be applied to all functions that access the Veros state object
      (even when subclassing :class:`veros.VerosSetup`).

    The first argument to the decorated function must be a VerosState instance.

    Veros routines cannot return anything. All changes must be applied to the passed state object.

    Parameters:
        dist_safe (bool): If set to False, all variables specified in local_variables are synced
            to the root process before execution and synced back after. This means that the routine
            will only be executed on rank 0. Has no effect in non-distributed contexts.

        local_variables (Tuple[str]): List of variable names to be synced if dist_safe=False. This
            must include all variables retrieved from the state object throughout the routine (inputs
            *and* outputs).

    Example:
       >>> from veros import VerosSetup, veros_routine
       >>>
       >>> class MyModel(VerosSetup):
       >>>     @veros_routine
       >>>     def set_topography(self, state):
       >>>         vs = state.variables
       >>>         settings = state.settings
       >>>         vs.kbot = npx.random.randint(0, settings.nz, size=vs.kbot.shape)

    """

    def inner_decorator(function):
        narg = 1 if _is_method(function) else 0
        num_params = len(inspect.signature(function).parameters)
        if narg >= num_params:
            raise TypeError("Veros routines must take at least one argument")

        routine = VerosRoutine(function, state_argnum=narg, dist_safe=dist_safe, local_variables=local_variables)
        routine = functools.wraps(function)(routine)
        return routine

    if function is not None:
        return inner_decorator(function)

    return inner_decorator


class VerosRoutine:
    """Do not instantiate directly!"""

    def __init__(self, function, dist_safe=True, local_variables=(), state_argnum=0):
        if isinstance(local_variables, str):
            local_variables = (local_variables,)

        self.function = function
        self.dist_safe = dist_safe
        self.local_variables = local_variables
        self.state_argnum = state_argnum
        self.name = _get_func_name(self.function)

    def __call__(self, *args, **kwargs):
        from veros import runtime_state as rst
        from veros.state import VerosState, DistSafeVariableWrapper
        from veros.core.operators import flush

        veros_state = args[self.state_argnum]

        if not isinstance(veros_state, VerosState):
            raise TypeError(f"Argument {self.state_argnum} to this Veros routine must be a VerosState object")

        timer = veros_state.profile_timers[self.name]

        with ExitStack() as es:
            vars_initialized = veros_state._variables is not None

            if vars_initialized:
                es.enter_context(veros_state.variables.unlock())

            execute = True
            restore_vars = False

            if not self.dist_safe:
                orig_vars = veros_state._variables
                if not isinstance(orig_vars, DistSafeVariableWrapper):
                    restore_vars = True
                    veros_state._variables = DistSafeVariableWrapper(orig_vars, self.local_variables)
                    veros_state._variables._gather_variables()

                execute = rst.proc_rank == 0

            routine_ctx = enter_routine(name=self.name, routine_obj=self, timer=timer, dist_safe=self.dist_safe)

            out = None
            try:
                with routine_ctx:
                    if execute:
                        out = self.function(*args, **kwargs)

            finally:
                if restore_vars:
                    veros_state._variables._scatter_variables()
                    veros_state._variables = orig_vars

                flush()

        if out is not None:
            logger.warning(
                f"Routine {self.name} returned object of type {type(out)}. Return objects are silently dropped."
            )

    def __get__(self, instance, _):
        return functools.partial(self.__call__, instance)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name} at {hex(id(self))}>"


# kernel


def veros_kernel(function=None, *, static_args=()):
    """Decorator that marks a function as a kernel that can be JIT compiled if supported
    by the backend.

    Kernels cannot modify the Veros state object. Instead, all modifications have to be
    returned explicitly.

    Parameters:
        static_args (Tuple[str]): Names of kernel arguments that should be static.

    Example:
        >>> from veros import veros_kernel, KernelOutput
        >>>
        >>> @veros_kernel
        >>> def double_psi(state):
        >>>     vs = state.variables
        >>>     vs.psi = 2 * vs.psi
        >>>     return KernelOutput(psi=vs.psi)

    """

    def inner_decorator(function):
        kernel = VerosKernel(function, static_args=static_args)
        kernel = functools.wraps(function)(kernel)
        return kernel

    if function is not None:
        return inner_decorator(function)

    return inner_decorator


class VerosKernel:
    """Do not instantiate directly!"""

    def __init__(self, function, static_args=()):
        """Do some parameter introspection."""

        # make sure function signature is in the form we need
        self.name = _get_func_name(function)
        self.func_sig = inspect.signature(function)

        func_params = self.func_sig.parameters

        allowed_param_types = (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)

        if any(p.kind not in allowed_param_types for p in func_params.values()):
            raise ValueError(f"Veros kernels do not support *args, **kwargs, or keyword-only parameters ({self.name})")

        # parse static args
        if isinstance(static_args, str):
            static_args = (static_args,)

        func_argnames = list(func_params.keys())

        self.static_argnums = []
        for static_arg in static_args:
            try:
                arg_index = func_argnames.index(static_arg)
            except ValueError:
                raise ValueError(
                    f'Veros kernel {self.name} has no argument "{static_arg}", but it is given in static_args'
                ) from None

            self.static_argnums.append(arg_index)

        self.function = function

    def __call__(self, *args, **kwargs):
        from veros import runtime_settings
        from veros.core.operators import flush

        # apply JIT
        if runtime_settings.backend == "jax":
            import jax

            CompiledFunction = type(jax.jit(lambda: None))

            if not isinstance(self.function, CompiledFunction):
                self.function = jax.jit(self.function, static_argnums=self.static_argnums)

        # JAX only accepts positional args when using static_argnums
        # so convert everything to positional for consistency
        bound_args = self.func_sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        veros_state = None
        for argval in bound_args.arguments.values():
            if isinstance(argval, VerosState):
                veros_state = argval
                break

        called_with_state = veros_state is not None

        # when profiling, make sure all inputs are ready before starting the timer
        if runtime_settings.profile_mode:
            flush()

        if called_with_state:
            timer = veros_state.profile_timers[self.name]
        else:
            timer = None

        with ExitStack() as es:
            if called_with_state:
                es.enter_context(veros_state.variables.unlock())

            args = list(bound_args.arguments.values())

            with enter_routine(self.name, self, timer):
                out = self.function(*args)

                if runtime_settings.profile_mode:
                    flush()

        return out

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name} at {hex(id(self))}>"


def is_veros_routine(func):
    if isinstance(func, functools.partial):
        func = func.func

    if inspect.ismethod(func):
        func = func.__self__

    return isinstance(func, VerosRoutine)



import functools

from veros import runtime_settings as rs, runtime_state as rst, logger


def memoize(func):
    func.cache = {}

    @functools.wraps(func)
    def inner(*args):
        if args not in func.cache:
            func.cache[args] = func(*args)

        return func.cache[args]

    return inner


def _get_solver_class():
    ls = rs.linear_solver

    def _get_best_solver():
        if rst.proc_num > 1:
            try:
                from veros.core.external.solvers.petsc_ import PETScSolver
            except ImportError:
                logger.warning("PETSc linear solver not available, falling back to SciPy")
            else:
                return PETScSolver

        if rs.backend == "jax" and rs.device == "gpu" and rs.float_type == "float64":
            from veros.core.external.solvers.scipy_jax import JAXSciPySolver

            return JAXSciPySolver

        from veros.core.external.solvers.scipy import SciPySolver

        return SciPySolver

    if ls == "best":
        return _get_best_solver()
    elif ls == "petsc":
        from veros.core.external.solvers.petsc_ import PETScSolver

        return PETScSolver
    elif ls == "scipy":
        from veros.core.external.solvers.scipy import SciPySolver

        return SciPySolver
    elif ls == "scipy_jax":
        from veros.core.external.solvers.scipy_jax import JAXSciPySolver

        return JAXSciPySolver

    raise ValueError(f"unrecognized linear solver {ls}")


@memoize
def get_linear_solver(state):
    logger.debug("Initializing linear solver")
    SolverClass = _get_solver_class()
    return SolverClass(state)



from veros.core.operators import numpy as npx

from veros import veros_routine, veros_kernel, KernelOutput
from veros.distributed import global_sum
from veros.variables import allocate
from veros.core import advection, diffusion, isoneutral, density, utilities
from veros.core.operators import update, update_add, at


@veros_kernel
def advect_tracer(state, tr):
    """
    calculate time tendency of a tracer due to advection
    """
    vs = state.variables
    settings = state.settings

    if settings.enable_superbee_advection:
        flux_east, flux_north, flux_top = advection.adv_flux_superbee(state, tr)
    else:
        flux_east, flux_north, flux_top = advection.adv_flux_2nd(state, tr)

    dtr = allocate(state.dimensions, ("xt", "yt", "zt"))
    dtr = update(
        dtr,
        at[2:-2, 2:-2, :],
        vs.maskT[2:-2, 2:-2, :]
        * (
            -(flux_east[2:-2, 2:-2, :] - flux_east[1:-3, 2:-2, :])
            / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
            - (flux_north[2:-2, 2:-2, :] - flux_north[2:-2, 1:-3, :])
            / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dyt[npx.newaxis, 2:-2, npx.newaxis])
        ),
    )
    dtr = update_add(dtr, at[:, :, 0], -1 * vs.maskT[:, :, 0] * flux_top[:, :, 0] / vs.dzt[0])
    dtr = update_add(
        dtr, at[:, :, 1:], -1 * vs.maskT[:, :, 1:] * (flux_top[:, :, 1:] - flux_top[:, :, :-1]) / vs.dzt[1:]
    )

    return dtr


@veros_kernel
def advect_temperature(state):
    """
    integrate temperature
    """
    vs = state.variables
    dtr = advect_tracer(state, vs.temp[..., vs.tau])
    vs.dtemp = update(vs.dtemp, at[..., vs.tau], dtr)
    return KernelOutput(dtemp=vs.dtemp)


@veros_kernel
def advect_salinity(state):
    """
    integrate salinity
    """
    vs = state.variables
    dtr = advect_tracer(state, vs.salt[..., vs.tau])
    vs.dsalt = update(vs.dsalt, at[..., vs.tau], dtr)
    return KernelOutput(dsalt=vs.dsalt)


@veros_kernel
def calc_eq_of_state(state, n):
    """
    calculate density, stability frequency, dynamic enthalpy and derivatives
    for time level n from temperature and salinity
    """
    vs = state.variables
    settings = state.settings

    salt = vs.salt[..., n]
    temp = vs.temp[..., n]
    press = npx.abs(vs.zt)

    """
    calculate new density
    """
    vs.rho = update(vs.rho, at[..., n], density.get_rho(state, salt, temp, press) * vs.maskT)

    """
    calculate new potential density
    """
    vs.prho = update(vs.prho, at[...], density.get_potential_rho(state, salt, temp) * vs.maskT)

    """
    calculate new dynamic enthalpy and derivatives
    """
    if settings.enable_conserve_energy:
        vs.Hd = update(vs.Hd, at[..., n], density.get_dyn_enthalpy(state, salt, temp, press) * vs.maskT)
        vs.int_drhodT = update(vs.int_drhodT, at[..., n], density.get_int_drhodT(state, salt, temp, press))
        vs.int_drhodS = update(vs.int_drhodS, at[..., n], density.get_int_drhodS(state, salt, temp, press))

    """
    new stability frequency
    """
    fxa = -settings.grav / settings.rho_0 / vs.dzw[npx.newaxis, npx.newaxis, :-1] * vs.maskW[:, :, :-1]
    vs.Nsqr = update(
        vs.Nsqr,
        at[:, :, :-1, n],
        fxa * (density.get_rho(state, salt[:, :, 1:], temp[:, :, 1:], press[:-1]) - vs.rho[:, :, :-1, n]),
    )
    vs.Nsqr = update(vs.Nsqr, at[:, :, -1, n], vs.Nsqr[:, :, -2, n])

    return KernelOutput(
        rho=vs.rho, prho=vs.prho, Hd=vs.Hd, int_drhodT=vs.int_drhodT, int_drhodS=vs.int_drhodS, Nsqr=vs.Nsqr
    )


@veros_kernel
def advect_temp_salt_enthalpy(state):
    """
    integrate temperature and salinity and diagnose sources of dynamic enthalpy
    """
    vs = state.variables
    settings = state.settings

    vs.dtemp = advect_temperature(state).dtemp
    vs.dsalt = advect_salinity(state).dsalt

    if settings.enable_conserve_energy:
        """
        advection of dynamic enthalpy
        """
        if settings.enable_superbee_advection:
            flux_east, flux_north, flux_top = advection.adv_flux_superbee(state, vs.Hd[:, :, :, vs.tau])
        else:
            flux_east, flux_north, flux_top = advection.adv_flux_2nd(state, vs.Hd[:, :, :, vs.tau])

        vs.dHd = update(
            vs.dHd,
            at[2:-2, 2:-2, :, vs.tau],
            vs.maskT[2:-2, 2:-2, :]
            * (
                -(flux_east[2:-2, 2:-2, :] - flux_east[1:-3, 2:-2, :])
                / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
                - (flux_north[2:-2, 2:-2, :] - flux_north[2:-2, 1:-3, :])
                / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dyt[npx.newaxis, 2:-2, npx.newaxis])
            ),
        )
        vs.dHd = update_add(vs.dHd, at[:, :, 0, vs.tau], -1 * vs.maskT[:, :, 0] * flux_top[:, :, 0] / vs.dzt[0])
        vs.dHd = update_add(
            vs.dHd,
            at[:, :, 1:, vs.tau],
            -1 * vs.maskT[:, :, 1:] * (flux_top[:, :, 1:] - flux_top[:, :, :-1]) / vs.dzt[npx.newaxis, npx.newaxis, 1:],
        )

        """
        changes in dyn. Enthalpy due to advection
        """
        diss = allocate(state.dimensions, ("xt", "yt", "zt"))
        diss = update(
            diss,
            at[2:-2, 2:-2, :],
            settings.grav
            / settings.rho_0
            * (
                -vs.int_drhodT[2:-2, 2:-2, :, vs.tau] * vs.dtemp[2:-2, 2:-2, :, vs.tau]
                - vs.int_drhodS[2:-2, 2:-2, :, vs.tau] * vs.dsalt[2:-2, 2:-2, :, vs.tau]
            )
            - vs.dHd[2:-2, 2:-2, :, vs.tau],
        )

        """
        contribution by vertical advection is - g rho w / rho0, substract this also
        """
        diss = update_add(
            diss,
            at[:, :, :-1],
            -0.25
            * settings.grav
            / settings.rho_0
            * vs.w[:, :, :-1, vs.tau]
            * (vs.rho[:, :, :-1, vs.tau] + vs.rho[:, :, 1:, vs.tau])
            * vs.dzw[npx.newaxis, npx.newaxis, :-1]
            / vs.dzt[npx.newaxis, npx.newaxis, :-1],
        )
        diss = update_add(
            diss,
            at[:, :, 1:],
            -0.25
            * settings.grav
            / settings.rho_0
            * vs.w[:, :, :-1, vs.tau]
            * (vs.rho[:, :, 1:, vs.tau] + vs.rho[:, :, :-1, vs.tau])
            * vs.dzw[npx.newaxis, npx.newaxis, :-1]
            / vs.dzt[npx.newaxis, npx.newaxis, 1:],
        )

    if settings.enable_conserve_energy and settings.enable_tke:
        """
        dissipation by advection interpolated on W-grid
        """
        vs.P_diss_adv = diffusion.dissipation_on_wgrid(state, diss, vs.kbot)

        """
        distribute P_diss_adv over domain, prevent draining of TKE
        """
        fxa = npx.sum(
            vs.area_t[2:-2, 2:-2, npx.newaxis]
            * vs.P_diss_adv[2:-2, 2:-2, :-1]
            * vs.dzw[npx.newaxis, npx.newaxis, :-1]
            * vs.maskW[2:-2, 2:-2, :-1]
        ) + npx.sum(0.5 * vs.area_t[2:-2, 2:-2] * vs.P_diss_adv[2:-2, 2:-2, -1] * vs.dzw[-1] * vs.maskW[2:-2, 2:-2, -1])

        tke_mask = vs.tke[2:-2, 2:-2, :-1, vs.tau] > 0.0

        fxb = npx.sum(
            vs.area_t[2:-2, 2:-2, npx.newaxis]
            * vs.dzw[npx.newaxis, npx.newaxis, :-1]
            * vs.maskW[2:-2, 2:-2, :-1]
            * tke_mask
        ) + npx.sum(0.5 * vs.area_t[2:-2, 2:-2] * vs.dzw[-1] * vs.maskW[2:-2, 2:-2, -1])

        fxa = global_sum(fxa)
        fxb = global_sum(fxb)

        vs.P_diss_adv = update(vs.P_diss_adv, at[2:-2, 2:-2, :-1], fxa / fxb * tke_mask)
        vs.P_diss_adv = update(vs.P_diss_adv, at[2:-2, 2:-2, -1], fxa / fxb)

    """
    Adam Bashforth time stepping for advection
    """
    vs.temp = update(
        vs.temp,
        at[:, :, :, vs.taup1],
        vs.temp[:, :, :, vs.tau]
        + settings.dt_tracer
        * ((1.5 + settings.AB_eps) * vs.dtemp[:, :, :, vs.tau] - (0.5 + settings.AB_eps) * vs.dtemp[:, :, :, vs.taum1])
        * vs.maskT,
    )
    vs.salt = update(
        vs.salt,
        at[:, :, :, vs.taup1],
        vs.salt[:, :, :, vs.tau]
        + settings.dt_tracer
        * ((1.5 + settings.AB_eps) * vs.dsalt[:, :, :, vs.tau] - (0.5 + settings.AB_eps) * vs.dsalt[:, :, :, vs.taum1])
        * vs.maskT,
    )

    return KernelOutput(
        temp=vs.temp, salt=vs.salt, dtemp=vs.dtemp, dsalt=vs.dsalt, dHd=vs.dHd, P_diss_adv=vs.P_diss_adv
    )


@veros_kernel
def vertmix_tempsalt(state):
    """
    vertical mixing of temperature and salinity
    """
    vs = state.variables
    settings = state.settings

    vs.dtemp_vmix = update(vs.dtemp_vmix, at[...], vs.temp[:, :, :, vs.taup1])
    vs.dsalt_vmix = update(vs.dsalt_vmix, at[...], vs.salt[:, :, :, vs.taup1])

    a_tri = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]
    b_tri = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]
    c_tri = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]
    d_tri = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]
    delta = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]

    _, water_mask, edge_mask = utilities.create_water_masks(vs.kbot[2:-2, 2:-2], settings.nz)

    delta = update(
        delta, at[:, :, :-1], settings.dt_tracer / vs.dzw[npx.newaxis, npx.newaxis, :-1] * vs.kappaH[2:-2, 2:-2, :-1]
    )
    delta = update(delta, at[:, :, -1], 0.0)
    a_tri = update(a_tri, at[:, :, 1:], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri = update(b_tri, at[:, :, 1:], 1 + (delta[:, :, 1:] + delta[:, :, :-1]) / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri_edge = 1 + delta / vs.dzt[npx.newaxis, npx.newaxis, :]
    c_tri = update(c_tri, at[:, :, :-1], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, :-1])
    d_tri = vs.temp[2:-2, 2:-2, :, vs.taup1]
    d_tri = update_add(d_tri, at[:, :, -1], settings.dt_tracer * vs.forc_temp_surface[2:-2, 2:-2] / vs.dzt[-1])

    sol = utilities.solve_implicit(a_tri, b_tri, c_tri, d_tri, water_mask, b_edge=b_tri_edge, edge_mask=edge_mask)
    vs.temp = update(vs.temp, at[2:-2, 2:-2, :, vs.taup1], npx.where(water_mask, sol, vs.temp[2:-2, 2:-2, :, vs.taup1]))

    d_tri = vs.salt[2:-2, 2:-2, :, vs.taup1]
    d_tri = update_add(d_tri, at[:, :, -1], settings.dt_tracer * vs.forc_salt_surface[2:-2, 2:-2] / vs.dzt[-1])

    sol = utilities.solve_implicit(a_tri, b_tri, c_tri, d_tri, water_mask, b_edge=b_tri_edge, edge_mask=edge_mask)
    vs.salt = update(vs.salt, at[2:-2, 2:-2, :, vs.taup1], npx.where(water_mask, sol, vs.salt[2:-2, 2:-2, :, vs.taup1]))

    vs.dtemp_vmix = (vs.temp[:, :, :, vs.taup1] - vs.dtemp_vmix) / settings.dt_tracer
    vs.dsalt_vmix = (vs.salt[:, :, :, vs.taup1] - vs.dsalt_vmix) / settings.dt_tracer

    """
    boundary exchange
    """
    vs.temp = update(
        vs.temp, at[..., vs.taup1], utilities.enforce_boundaries(vs.temp[..., vs.taup1], settings.enable_cyclic_x)
    )
    vs.salt = update(
        vs.salt, at[..., vs.taup1], utilities.enforce_boundaries(vs.salt[..., vs.taup1], settings.enable_cyclic_x)
    )

    return KernelOutput(dtemp_vmix=vs.dtemp_vmix, temp=vs.temp, dsalt_vmix=vs.dsalt_vmix, salt=vs.salt)


@veros_kernel
def surf_densityf(state):
    """
    surface density flux
    """
    vs = state.variables

    vs.forc_rho_surface = vs.maskT[:, :, -1] * (
        density.get_drhodT(state, vs.salt[:, :, -1, vs.taup1], vs.temp[:, :, -1, vs.taup1], npx.abs(vs.zt[-1]))
        * vs.forc_temp_surface
        + density.get_drhodS(state, vs.salt[:, :, -1, vs.taup1], vs.temp[:, :, -1, vs.taup1], npx.abs(vs.zt[-1]))
        * vs.forc_salt_surface
    )

    return KernelOutput(forc_rho_surface=vs.forc_rho_surface)


@veros_kernel
def diag_P_diss_v(state):
    vs = state.variables
    settings = state.settings

    vs.P_diss_v = update(vs.P_diss_v, at[...], 0.0)
    aloc = allocate(state.dimensions, ("xt", "yt", "zt"))

    if settings.enable_conserve_energy:
        """
        diagnose dissipation of dynamic enthalpy by vertical mixing
        """
        fxa = (-vs.int_drhodT[2:-2, 2:-2, 1:, vs.taup1] + vs.int_drhodT[2:-2, 2:-2, :-1, vs.taup1]) / vs.dzw[
            npx.newaxis, npx.newaxis, :-1
        ]
        vs.P_diss_v = update_add(
            vs.P_diss_v,
            at[2:-2, 2:-2, :-1],
            -settings.grav
            / settings.rho_0
            * fxa
            * vs.kappaH[2:-2, 2:-2, :-1]
            * (vs.temp[2:-2, 2:-2, 1:, vs.taup1] - vs.temp[2:-2, 2:-2, :-1, vs.taup1])
            / vs.dzw[npx.newaxis, npx.newaxis, :-1]
            * vs.maskW[2:-2, 2:-2, :-1],
        )
        fxa = (-vs.int_drhodS[2:-2, 2:-2, 1:, vs.taup1] + vs.int_drhodS[2:-2, 2:-2, :-1, vs.taup1]) / vs.dzw[
            npx.newaxis, npx.newaxis, :-1
        ]
        vs.P_diss_v = update_add(
            vs.P_diss_v,
            at[2:-2, 2:-2, :-1],
            -settings.grav
            / settings.rho_0
            * fxa
            * vs.kappaH[2:-2, 2:-2, :-1]
            * (vs.salt[2:-2, 2:-2, 1:, vs.taup1] - vs.salt[2:-2, 2:-2, :-1, vs.taup1])
            / vs.dzw[npx.newaxis, npx.newaxis, :-1]
            * vs.maskW[2:-2, 2:-2, :-1],
        )

        fxa = 2 * vs.int_drhodT[2:-2, 2:-2, -1, vs.taup1] / vs.dzw[-1]
        vs.P_diss_v = update_add(
            vs.P_diss_v,
            at[2:-2, 2:-2, -1],
            -settings.grav / settings.rho_0 * fxa * vs.forc_temp_surface[2:-2, 2:-2] * vs.maskW[2:-2, 2:-2, -1],
        )
        fxa = 2 * vs.int_drhodS[2:-2, 2:-2, -1, vs.taup1] / vs.dzw[-1]
        vs.P_diss_v = update_add(
            vs.P_diss_v,
            at[2:-2, 2:-2, -1],
            -settings.grav / settings.rho_0 * fxa * vs.forc_salt_surface[2:-2, 2:-2] * vs.maskW[2:-2, 2:-2, -1],
        )

    if settings.enable_conserve_energy:
        """
        determine effect due to nonlinear equation of state
        """
        aloc = update(aloc, at[:, :, :-1], vs.kappaH[:, :, :-1] * vs.Nsqr[:, :, :-1, vs.taup1])
        vs.P_diss_nonlin = update(vs.P_diss_nonlin, at[:, :, :-1], vs.P_diss_v[:, :, :-1] - aloc[:, :, :-1])
        vs.P_diss_v = update(vs.P_diss_v, at[:, :, :-1], aloc[:, :, :-1])
    else:
        """
        diagnose N^2 vs. kappaH, i.e. exchange of pot. energy with TKE
        """
        vs.P_diss_v = update(vs.P_diss_v, at[:, :, :-1], vs.kappaH[:, :, :-1] * vs.Nsqr[:, :, :-1, vs.taup1])
        vs.P_diss_v = update(
            vs.P_diss_v, at[:, :, -1], -vs.forc_rho_surface * vs.maskT[:, :, -1] * settings.grav / settings.rho_0
        )

    return KernelOutput(P_diss_v=vs.P_diss_v, P_diss_nonlin=vs.P_diss_nonlin)


@veros_routine
def thermodynamics(state):
    """
    integrate temperature and salinity and diagnose sources of dynamic enthalpy
    """
    """
    Advection tendencies for temperature, salinity and dynamic enthalpy
    """
    vs = state.variables
    settings = state.settings

    vs.update(advect_temp_salt_enthalpy(state))

    """
    horizontal diffusion
    """
    with state.timers["isoneutral"]:
        if settings.enable_hor_diffusion:
            vs.update(diffusion.tempsalt_diffusion(state))

        if settings.enable_biharmonic_mixing:
            vs.update(diffusion.tempsalt_biharmonic(state))

        """
        sources like restoring zones, etc
        """
        if settings.enable_tempsalt_sources:
            vs.update(diffusion.tempsalt_sources(state))

        """
        isopycnal diffusion
        """
        if settings.enable_neutral_diffusion:
            vs.P_diss_iso = update(vs.P_diss_iso, at[...], 0.0)
            vs.dtemp_iso = update(vs.dtemp_iso, at[...], 0.0)
            vs.dsalt_iso = update(vs.dsalt_iso, at[...], 0.0)

            vs.update(isoneutral.isoneutral_diffusion_pre(state))
            vs.update(isoneutral.isoneutral_diffusion(state, tr=vs.temp, istemp=True))
            vs.update(isoneutral.isoneutral_diffusion(state, tr=vs.salt, istemp=False))

            if settings.enable_skew_diffusion:
                vs.P_diss_skew = update(vs.P_diss_skew, at[...], 0.0)
                vs.update(isoneutral.isoneutral_skew_diffusion(state, tr=vs.temp, istemp=True))
                vs.update(isoneutral.isoneutral_skew_diffusion(state, tr=vs.salt, istemp=False))

    with state.timers["vmix"]:
        vs.update(vertmix_tempsalt(state))

    with state.timers["eq_of_state"]:
        vs.update(calc_eq_of_state(state, vs.taup1))

    """
    surface density flux
    """
    vs.update(surf_densityf(state))

    with state.timers["vmix"]:
        vs.update(diag_P_diss_v(state))



from veros import logger

from veros.variables import Variable
from veros.core.operators import numpy as npx
from veros.diagnostics.base import VerosDiagnostic
from veros.distributed import global_sum


class TracerMonitor(VerosDiagnostic):
    """Diagnostic monitoring global tracer contents / fluxes.

    Writes output to stdout (no binary output).
    """

    name = "tracer_monitor"
    output_frequency = None

    def __init__(self, state):
        self.var_meta = {
            "tempm1": Variable("tempm1", None, write_to_restart=True),
            "vtemp1": Variable("vtemp1", None, write_to_restart=True),
            "saltm1": Variable("saltm1", None, write_to_restart=True),
            "vsalt1": Variable("vsalt1", None, write_to_restart=True),
        }

    def initialize(self, state):
        self.initialize_variables(state)

    def diagnose(self, state):
        pass

    def output(self, state):
        """
        Diagnose tracer content
        """
        vs = state.variables
        tracer_vs = self.variables

        cell_volume = vs.area_t[2:-2, 2:-2, npx.newaxis] * vs.dzt[npx.newaxis, npx.newaxis, :] * vs.maskT[2:-2, 2:-2, :]
        volm = global_sum(npx.sum(cell_volume))
        tempm = global_sum(npx.sum(cell_volume * vs.temp[2:-2, 2:-2, :, vs.tau]))
        saltm = global_sum(npx.sum(cell_volume * vs.salt[2:-2, 2:-2, :, vs.tau]))
        vtemp = global_sum(npx.sum(cell_volume * vs.temp[2:-2, 2:-2, :, vs.tau] ** 2))
        vsalt = global_sum(npx.sum(cell_volume * vs.salt[2:-2, 2:-2, :, vs.tau] ** 2))

        logger.diagnostic(
            f" Mean temperature {tempm / volm:.2e} change to last {(tempm - tracer_vs.tempm1) / volm:.2e}"
        )
        logger.diagnostic(
            f" Mean salinity    {saltm / volm:.2e} change to last {(saltm - tracer_vs.saltm1) / volm:.2e}"
        )
        logger.diagnostic(
            f" Temperature var. {vtemp / volm:.2e} change to last {(vtemp - tracer_vs.vtemp1) / volm:.2e}"
        )
        logger.diagnostic(
            f" Salinity var.    {vsalt / volm:.2e} change to last {(vsalt - tracer_vs.vsalt1) / volm:.2e}"
        )

        tracer_vs.tempm1 = tempm
        tracer_vs.vtemp1 = vtemp
        tracer_vs.saltm1 = saltm
        tracer_vs.vsalt1 = vsalt



from veros import veros_kernel
from veros.core.density import gsw, linear_eq as lq, nonlinear_eq1 as nq1, nonlinear_eq2 as nq2, nonlinear_eq3 as nq3


@veros_kernel
def get_rho(state, salt_loc, temp_loc, press):
    """
    calculate density as a function of temperature, salinity and pressure
    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return lq.linear_eq_of_state_rho(salt_loc, temp_loc)
    elif settings.eq_of_state_type == 2:
        return nq1.nonlin1_eq_of_state_rho(salt_loc, temp_loc)
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_rho(salt_loc, temp_loc, press)
    elif settings.eq_of_state_type == 4:
        return nq3.nonlin3_eq_of_state_rho(salt_loc, temp_loc)
    elif settings.eq_of_state_type == 5:
        return gsw.gsw_rho(salt_loc, temp_loc, press)
    else:
        raise ValueError("unknown equation of state")


@veros_kernel
def get_potential_rho(state, salt_loc, temp_loc, press_ref=0.0):
    """
    calculate potential density as a function of temperature, salinity
    and reference pressure

    Note:

        This is identical to get_rho for eq_of_state_type {1, 2, 4}

    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return lq.linear_eq_of_state_rho(salt_loc, temp_loc)
    elif settings.eq_of_state_type == 2:
        return nq1.nonlin1_eq_of_state_rho(salt_loc, temp_loc)
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_rho(salt_loc, temp_loc, press_ref)
    elif settings.eq_of_state_type == 4:
        return nq3.nonlin3_eq_of_state_rho(salt_loc, temp_loc)
    elif settings.eq_of_state_type == 5:
        return gsw.gsw_rho(salt_loc, temp_loc, press_ref)
    else:
        raise ValueError("unknown equation of state")


@veros_kernel
def get_dyn_enthalpy(state, salt_loc, temp_loc, press):
    """
    calculate dynamic enthalpy as a function of temperature, salinity and pressure
    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return lq.linear_eq_of_state_dyn_enthalpy(salt_loc, temp_loc, press)
    elif settings.eq_of_state_type == 2:
        return nq1.nonlin1_eq_of_state_dyn_enthalpy(salt_loc, temp_loc, press)
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_dyn_enthalpy(salt_loc, temp_loc, press)
    elif settings.eq_of_state_type == 4:
        return nq3.nonlin3_eq_of_state_dyn_enthalpy(salt_loc, temp_loc, press)
    elif settings.eq_of_state_type == 5:
        return gsw.gsw_dyn_enthalpy(salt_loc, temp_loc, press)
    else:
        raise ValueError("unknown equation of state")


@veros_kernel
def get_salt(state, rho_loc, temp_loc, press_loc):
    """
    calculate salinity as a function of density, temperature and pressure
    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return lq.linear_eq_of_state_salt(rho_loc, temp_loc)
    elif settings.eq_of_state_type == 2:
        return nq1.nonlin1_eq_of_state_salt(rho_loc, temp_loc)
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_salt(rho_loc, temp_loc, press_loc)
    elif settings.eq_of_state_type == 4:
        return nq3.nonlin3_eq_of_state_salt(rho_loc, temp_loc)
    else:
        raise ValueError("unknown equation of state")


@veros_kernel
def get_drhodT(state, salt_loc, temp_loc, press_loc):
    """
    calculate drho/dT as a function of temperature, salinity and pressure
    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return lq.linear_eq_of_state_drhodT()
    elif settings.eq_of_state_type == 2:
        return nq1.nonlin1_eq_of_state_drhodT(temp_loc)
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_drhodT(temp_loc, press_loc)
    elif settings.eq_of_state_type == 4:
        return nq3.nonlin3_eq_of_state_drhodT(temp_loc)
    elif settings.eq_of_state_type == 5:
        return gsw.gsw_drhodT(salt_loc, temp_loc, press_loc)
    else:
        raise ValueError("unknown equation of state")


@veros_kernel
def get_drhodS(state, salt_loc, temp_loc, press_loc):
    """
    calculate drho/dS as a function of temperature, salinity and pressure
    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return lq.linear_eq_of_state_drhodS()
    elif settings.eq_of_state_type == 2:
        return nq1.nonlin1_eq_of_state_drhodS()
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_drhodS()
    elif settings.eq_of_state_type == 4:
        return nq3.nonlin3_eq_of_state_drhodS()
    elif settings.eq_of_state_type == 5:
        return gsw.gsw_drhodS(salt_loc, temp_loc, press_loc)
    else:
        raise ValueError("unknown equation of state")


@veros_kernel
def get_drhodp(state, salt_loc, temp_loc, press_loc):
    """
    calculate drho/dP as a function of temperature, salinity and pressure
    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return lq.linear_eq_of_state_drhodp()
    elif settings.eq_of_state_type == 2:
        return nq1.nonlin1_eq_of_state_drhodp()
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_drhodp(temp_loc)
    elif settings.eq_of_state_type == 4:
        return nq3.nonlin3_eq_of_state_drhodp()
    elif settings.eq_of_state_type == 5:
        return gsw.gsw_drhodp(salt_loc, temp_loc, press_loc)
    else:
        raise ValueError("unknown equation of state")


@veros_kernel
def get_int_drhodT(state, salt_loc, temp_loc, press_loc):
    """
    calculate int_z^0 drho/dT dz' as a function of temperature, salinity and pressure
    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return press_loc * lq.linear_eq_of_state_drhodT()  # int_z^0rho_T dz = - rho_T z
    elif settings.eq_of_state_type == 2:
        return press_loc * nq1.nonlin1_eq_of_state_drhodT(temp_loc)
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_int_drhodT(temp_loc, press_loc)
    elif settings.eq_of_state_type == 4:
        return press_loc * nq3.nonlin3_eq_of_state_drhodT(temp_loc)
    elif settings.eq_of_state_type == 5:
        return -(1024.0 / 9.81) * gsw.gsw_dHdT(salt_loc, temp_loc, press_loc)
    else:
        raise ValueError("unknown equation of state")


@veros_kernel
def get_int_drhodS(state, salt_loc, temp_loc, press_loc):
    """
    calculate int_z^0 drho/dS dz' as a function of temperature, salinity and pressure
    """
    settings = state.settings

    if settings.eq_of_state_type == 1:
        return press_loc * lq.linear_eq_of_state_drhodS()  # int_z^0rho_T dz = - rho_T z
    elif settings.eq_of_state_type == 2:
        return press_loc * nq1.nonlin1_eq_of_state_drhodS()
    elif settings.eq_of_state_type == 3:
        return nq2.nonlin2_eq_of_state_int_drhodS(press_loc)
    elif settings.eq_of_state_type == 4:
        return press_loc * nq3.nonlin3_eq_of_state_drhodS()
    elif settings.eq_of_state_type == 5:
        return -(1024.0 / 9.81) * gsw.gsw_dHdS(salt_loc, temp_loc, press_loc)
    else:
        raise ValueError("unknown equation of state")



import timeit
import threading

timer_context = threading.local()
timer_context.active = True


class Timer:
    def __init__(self):
        self.total_time = 0
        self.last_time = 0

    def __enter__(self):
        self.start_time = timeit.default_timer()

    def __exit__(self, *args, **kwargs):
        self.last_time = timeit.default_timer() - self.start_time

        if timer_context.active:
            self.total_time += self.last_time



#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages
from setuptools.extension import Extension

from codecs import open
import os
import re
import sys

from Cython.Build import cythonize

here = os.path.abspath(os.path.dirname(__file__))
sys.path.append(here)
import versioneer  # noqa: E402
import cuda_ext  # noqa: E402


CLASSIFIERS = """
Development Status :: 4 - Beta
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Programming Language :: Python :: 3
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3.11
Programming Language :: Python :: 3.12
Programming Language :: Python :: Implementation :: CPython
Topic :: Scientific/Engineering
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

MINIMUM_VERSIONS = {
    "numpy": "1.13",
    "requests": "2.18",
    "jax": "0.2.10",
}


CONSOLE_SCRIPTS = [
    "veros = veros.cli.veros:cli",
    "veros-run = veros.cli.veros_run:cli",
    "veros-copy-setup = veros.cli.veros_copy_setup:cli",
    "veros-resubmit = veros.cli.veros_resubmit:cli",
    "veros-create-mask = veros.cli.veros_create_mask:cli",
]

PACKAGE_DATA = ["setups/*/assets.json", "setups/*/*.npy", "setups/*/*.png"]

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def parse_requirements(reqfile):
    requirements = []

    with open(os.path.join(here, reqfile), encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            parsed = re.match(r"(\w+)(.*?)(;.*)?$", line)
            pkg, deps, extra = parsed.groups()
            if extra is None:
                extra = ""
            deps = deps.replace("==", "<=")
            if pkg in MINIMUM_VERSIONS:
                deps = f"{deps},>={MINIMUM_VERSIONS[pkg]}"
            line = "".join([pkg, deps, extra])
            requirements.append(line)

    return requirements


INSTALL_REQUIRES = parse_requirements("requirements.txt")

jax_req = parse_requirements("requirements_jax.txt")
for line in jax_req:  # inject jaxlib requirement
    if line.startswith("jax"):
        jax_req.append(line.replace("jax", "jaxlib"))
        break

EXTRAS_REQUIRE = {
    "test": ["pytest", "pytest-cov", "pytest-forked", "xarray"],
    "jax": jax_req,
}


def get_extensions(require_cython_ext, require_cuda_ext):
    cuda_info = cuda_ext.cuda_info

    extension_modules = {
        "veros.core.special.tdma_cython_": ["tdma_cython_.pyx"],
        "veros.core.special.tdma_cuda_": ["tdma_cuda_.pyx", "cuda_tdma_kernels.cu"],
    }

    def is_cuda_ext(sources):
        return any(source.endswith(".cu") for source in sources)

    extensions = []
    for module, sources in extension_modules.items():
        extension_dir = os.path.join(*module.split(".")[:-1])

        kwargs = dict()
        if is_cuda_ext(sources):
            kwargs.update(
                library_dirs=cuda_info["lib64"],
                libraries=["cudart"],
                runtime_library_dirs=cuda_info["lib64"],
                include_dirs=cuda_info["include"],
            )

        ext = Extension(
            name=module,
            sources=[os.path.join(extension_dir, f) for f in sources],
            extra_compile_args={
                "gcc": [],
                "nvcc": cuda_info["cflags"],
            },
            **kwargs,
        )

        extensions.append(ext)

    extensions = cythonize(extensions, language_level=3, exclude_failures=True)

    for ext in extensions:
        is_required = (not is_cuda_ext(ext.sources) and require_cython_ext) or (
            is_cuda_ext(ext.sources) and require_cuda_ext
        )

        if not is_required:
            ext.optional = True

    return extensions


cmdclass = versioneer.get_cmdclass()
build_ext = type("custom_build_ext", (cuda_ext.custom_build_ext, cmdclass["build_ext"]), {})
cmdclass.update(build_ext=build_ext)


def _env_to_bool(envvar):
    return os.environ.get(envvar, "").lower() in ("1", "true", "on")


extensions = get_extensions(
    require_cython_ext=_env_to_bool("VEROS_REQUIRE_CYTHON_EXT"),
    require_cuda_ext=_env_to_bool("VEROS_REQUIRE_CUDA_EXT"),
)

setup(
    name="veros",
    license="MIT",
    author="Dion Häfner (NBI Copenhagen)",
    author_email="dion.haefner@nbi.ku.dk",
    keywords="oceanography python parallel numpy multi-core geophysics ocean-model mpi4py jax",
    description="The versatile ocean simulator, in pure Python, powered by JAX.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://veros.readthedocs.io",
    python_requires=">=3.10",
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    ext_modules=extensions,
    entry_points={"console_scripts": CONSOLE_SCRIPTS, "veros.setup_dirs": ["base = veros.setups"]},
    package_data={"veros": PACKAGE_DATA},
    classifiers=[c for c in CLASSIFIERS.split("\n") if c],
    zip_safe=False,
)



import threading
import contextlib

from veros import logger, runtime_settings, runtime_state


@contextlib.contextmanager
def threaded_io(filepath, mode):
    """
    If using IO threads, start a new thread to write the HDF5 data to disk.
    """
    import h5py

    if runtime_settings.use_io_threads:
        _wait_for_disk(filepath)
        _io_locks[filepath].clear()
    kwargs = {}
    if runtime_state.proc_num > 1:
        kwargs.update(driver="mpio", comm=runtime_settings.mpi_comm)
    h5file = h5py.File(filepath, mode, **kwargs)
    try:
        yield h5file
    finally:
        if runtime_settings.use_io_threads:
            threading.Thread(target=_write_to_disk, args=(h5file, filepath)).start()
        else:
            _write_to_disk(h5file, filepath)


_io_locks = {}


def _add_to_locks(file_id):
    """
    If there is no lock for file_id, create one
    """
    if file_id not in _io_locks:
        _io_locks[file_id] = threading.Event()
        _io_locks[file_id].set()


def _wait_for_disk(file_id):
    """
    Wait for the lock of file_id to be released
    """
    logger.debug(f"Waiting for lock {file_id} to be released")
    _add_to_locks(file_id)
    lock_released = _io_locks[file_id].wait(runtime_settings.io_timeout)
    if not lock_released:
        raise RuntimeError("Timeout while waiting for disk IO to finish")


def _write_to_disk(h5file, file_id):
    """
    Sync HDF5 data to disk, close file handle, and release lock.
    May run in a separate thread.
    """
    try:
        h5file.close()
    finally:
        if runtime_settings.use_io_threads and file_id is not None:
            _io_locks[file_id].set()
